# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2022 Pietro Albini <pietro.albini@ferrous-systems.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Click code for spdx subcommand."""

import contextlib
import logging
import sys

import click

from ..covered_files import _IGNORE_SPDX_PATTERNS
from ..i18n import _
from ..report import ProjectReport
from .common import ClickObj
from .main import main

_LOGGER = logging.getLogger(__name__)

_HELP = _("Generate an SPDX bill of materials.")


@main.command(name="spdx", help=_HELP)
@click.option(
    "--output",
    "-o",
    type=click.File("w", encoding="utf-8", lazy=True),
    # Default is stdout.
    default=None,
    help=_("File to write to."),
)
@click.option(
    "--add-license-concluded",
    is_flag=True,
    help=_(
        "Populate the LicenseConcluded field; note that reuse cannot guarantee"
        " that the field is accurate."
    ),
)
@click.option(
    "--add-licence-concluded",
    "add_license_concluded",
    hidden=True,
)
@click.option(
    "--creator-person",
    type=str,
    help=_("Name of the person signing off on the SPDX report."),
)
@click.option(
    "--creator-organization",
    help=_("Name of the organization signing off on the SPDX report."),
)
@click.option(
    "--creator-organisation",
    "creator_organization",
    hidden=True,
)
@click.pass_obj
def spdx(
    obj: ClickObj,
    output: click.File | None,
    add_license_concluded: bool,
    creator_person: str | None,
    creator_organization: str | None,
) -> None:
    # pylint: disable=missing-function-docstring

    # The SPDX spec mandates that a creator must be specified when a license
    # conclusion is made, so here we enforce that. More context:
    # https://github.com/fsfe/reuse-tool/issues/586#issuecomment-1310425706
    if (
        add_license_concluded
        and creator_person is None
        and creator_organization is None
    ):
        raise click.UsageError(
            _(
                "'--creator-person' or '--creator-organization'"
                " is required when '--add-license-concluded' is provided."
            )
        )

    if (
        output is not None
        and output.name != "-"
        and not any(
            pattern.match(output.name) for pattern in _IGNORE_SPDX_PATTERNS
        )
    ):
        # pylint: disable=line-too-long
        _LOGGER.warning(
            _(
                "'{path}' does not match a common SPDX file pattern. Find"
                " the suggested naming conventions here:"
                " https://spdx.github.io/spdx-spec/conformance/#44-standard-data-format-requirements"
            ).format(path=output.name)
        )

    report = ProjectReport.generate(
        obj.project,
        multiprocessing=not obj.no_multiprocessing,
        add_license_concluded=add_license_concluded,
    )

    with contextlib.ExitStack() as stack:
        if output is not None:
            out = stack.enter_context(output.open())  # type: ignore
        else:
            out = sys.stdout
        click.echo(
            report.bill_of_materials(
                creator_person=creator_person,
                creator_organization=creator_organization,
            ),
            file=out,
        )



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2021 Alliander N.V.
# SPDX-FileCopyrightText: 2023 Carmen Bianca BAKKER <carmenbianca@fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Utilities related to the parsing and storing of copyright notices."""

import difflib
import logging
import re
from collections import Counter, defaultdict
from collections.abc import Iterable
from dataclasses import InitVar, dataclass, field
from enum import Enum, unique
from functools import cached_property
from io import StringIO
from itertools import chain
from typing import Any, Literal, NewType, cast

from license_expression import (
    ExpressionError,
    LicenseExpression,
    Licensing,
    combine_expressions,
)

from .exceptions import CopyrightNoticeParseError, YearRangeParseError

# REUSE-IgnoreStart

_LOGGER = logging.getLogger(__name__)

_LICENSING = Licensing()

#: A string that is four digits long.
FourDigitString = NewType("FourDigitString", str)
#: A range separator between two years.
YearRangeSeparator = Literal[
    "--",  # ascii en dash
    "–",  # en dash
    "-",  # ascii dash
]


def is_four_digits(value: str) -> FourDigitString | Literal[False]:
    """Identify a string as a four-digit string. Return the string as
    :type:`FourDigitString` if it is one.

    >>> is_four_digits("1234")
    '1234'
    >>> is_four_digits("abcd")
    False
    >>> is_four_digits("12345")
    False
    """
    if value.isdigit() and len(value) == 4:
        return FourDigitString(value)
    return False


def validate_four_digits(value: str) -> FourDigitString:
    """Validate whether a given string is a :type:`FourDigitString`.

    >>> validate_four_digits("1234")
    '1234'
    >>> validate_four_digits("abcd")
    Traceback (most recent call last):
        ...
    ValueError: 'abcd' is not a four-digit year.
    >>> validate_four_digits("12345")
    Traceback (most recent call last):
        ...
    ValueError: '12345' is not a four-digit year.

    Raises:
        ValueError: The string is not four digits.
    """
    if not (result := is_four_digits(value)):
        raise ValueError(f"'{value}' is not a four-digit year.")
    return result


_ANY_SEPARATOR = (
    r"(?:" + r"|".join(YearRangeSeparator.__args__) + r")"  # type: ignore
)
#: A regex pattern to match e.g. '2017-2020'.
YEAR_RANGE_PATTERN = re.compile(
    r"(?P<start>\d{4})"
    r"(?:"
    r"(?:(?P<separator_nonspaced>" + _ANY_SEPARATOR + r")"
    r"(?P<end_nonspaced>\S+)?)"
    r"|"
    r"(?:\s+(?P<separator_spaced>" + _ANY_SEPARATOR + r")"
    r"\s+(?P<end_spaced>\d{4}))"
    r")?"
)
_YEAR_RANGE_PATTERN_ANONYMISED = re.sub(
    r"\(\?P<\w+>", r"(?:", YEAR_RANGE_PATTERN.pattern
)
_SYMBOL_OR_C_SUBPATTERN = r"(©|\([Cc]\))"
_STRING_SUBPATTERN = (
    r"(Copyright((\s*" + _SYMBOL_OR_C_SUBPATTERN + r")|(?=\s)))"
)
#: A regex pattern to match a complete and valid REUSE copyright notice.
COPYRIGHT_NOTICE_PATTERN = re.compile(
    r"(?P<prefix>("
    r"SPDX-(File|Snippet)CopyrightText:"
    + r"(\s*("
    + _SYMBOL_OR_C_SUBPATTERN
    + "|"
    + _STRING_SUBPATTERN
    + "))?"
    + r"|"
    + _STRING_SUBPATTERN
    + r"|"
    r"©"
    r"))"
    r"\s*"
    r"(?P<text>.*?)"
    r"\s*"
)
_YEARS_PATTERN = re.compile(
    r"(?P<prefix>(^|,?\s+))(?P<years>"
    + _YEAR_RANGE_PATTERN_ANONYMISED
    + r"((\s*,\s*|\s+)"
    + _YEAR_RANGE_PATTERN_ANONYMISED
    + r")*)(?P<suffix>,?(\s+|$))"
)
_COMMA_SPACE_PATTERN = re.compile(r"^,?\s+")

_LOOKBEHINDS = "".join(
    rf"(?<!\s{separator})"
    for separator in YearRangeSeparator.__args__  # type: ignore
)
_YEAR_RANGE_SPLIT_REGEX = re.compile(
    # Separated by comma (plus any whitespace)
    r",\s*|" +
    # Separated by whitespace only. However, we cannot split on
    # whitespace that is itself part of a year range (e.g. ``2017 -
    # 2019``). The lookbehinds and lookahead take care of that. There
    # are multiple lookbehinds because lookbehinds cannot be
    # variable-width.
    _LOOKBEHINDS + r"\s+"
    rf"(?!{_ANY_SEPARATOR}\s)"
)


# TODO: In Python 3.11, turn this into a StrEnum
@unique
class CopyrightPrefix(Enum):
    """The prefix used for a copyright notice."""

    SPDX = "SPDX-FileCopyrightText:"
    SPDX_C = "SPDX-FileCopyrightText: (C)"
    SPDX_SYMBOL = "SPDX-FileCopyrightText: ©"
    SPDX_STRING = "SPDX-FileCopyrightText: Copyright"
    SPDX_STRING_C = "SPDX-FileCopyrightText: Copyright (C)"
    SPDX_STRING_SYMBOL = "SPDX-FileCopyrightText: Copyright ©"
    SNIPPET = "SPDX-SnippetCopyrightText:"
    SNIPPET_C = "SPDX-SnippetCopyrightText: (C)"
    SNIPPET_SYMBOL = "SPDX-SnippetCopyrightText: ©"
    SNIPPET_STRING = "SPDX-SnippetCopyrightText: Copyright"
    SNIPPET_STRING_C = "SPDX-SnippetCopyrightText: Copyright (C)"
    SNIPPET_STRING_SYMBOL = "SPDX-SnippetCopyrightText: Copyright ©"
    STRING = "Copyright"
    STRING_C = "Copyright (C)"
    STRING_SYMBOL = "Copyright ©"
    SYMBOL = "©"

    @staticmethod
    def lowercase_name(name: str) -> str:
        """Given an uppercase NAME, return name. Underscores are converted to
        dashes.

        >>> CopyrightPrefix.lowercase_name("SPDX_STRING")
        'spdx-string'
        """
        return name.lower().replace("_", "-")

    @staticmethod
    def uppercase_name(name: str) -> str:
        """Given a lowercase name, return NAME. Dashes are converted to
        underscores.

        >>> CopyrightPrefix.uppercase_name("spdx-string")
        'SPDX_STRING'
        """
        return name.upper().replace("-", "_")


@dataclass(frozen=True)
class YearRange:
    """Represents a year range, such as '2017-2025', or '2017'. This only
    represents a single range; multiple separated ranges should be put in a
    collection (typically a tuple).
    """

    #: The first year in the range. If it is only a single year, this is the
    #: only relevant value.
    start: FourDigitString
    #: The separator between :attr:`start` and :attr:`end`. If no value for
    #: :attr:`end` is provided, a range into infinity is implied, and
    #: :attr:`end` becomes an empty string.
    separator: YearRangeSeparator | None = field(default=None, compare=False)
    #: The second year in the range. This can also be a word like 'Present'.
    #: This is bad practice, but still supported.
    end: FourDigitString | str | None = None

    #: If parsed from a string, this contains the original string.
    original: str | None = field(
        default=None, init=False, repr=False, compare=False
    )

    def __post_init__(self) -> None:
        if self.separator is not None and self.end is None:
            object.__setattr__(self, "end", "")

    # TODO: In Python 3.11, return Self
    @classmethod
    def from_string(cls, value: str) -> "YearRange":
        """Create a :class:`YearRange` object from a string.

        Raises:
            YearRangeParseError: The string is not a valid year range.
        """
        re_result = YEAR_RANGE_PATTERN.fullmatch(value)
        if not re_result:
            raise YearRangeParseError(f"'{value}' is not a valid year range.")

        groups = re_result.groupdict()
        start = groups["start"]
        separator = groups["separator_nonspaced"] or groups["separator_spaced"]
        end = groups["end_nonspaced"] or groups["end_spaced"]

        # Mypy is disabled for this because the values are enforced by the
        # regex. This could be cleaner, but would require a lot of useless code
        # to validate what the regex already enforces.
        result = cls(start, separator, end)  # type: ignore
        object.__setattr__(result, "original", value)
        return result

    @classmethod
    def tuple_from_string(cls, value: str) -> tuple["YearRange", ...]:
        """Create a tuple of :class:`YearRange` objects from a string containing
        multiple ranges.

        Raises:
            YearRangeParseError: The substring is not a valid year range.
        """
        years: list[YearRange] = []
        for year in _YEAR_RANGE_SPLIT_REGEX.split(value):
            if not year:
                continue
            years.append(YearRange.from_string(year))
        return tuple(years)

    def __str__(self) -> str:
        result = StringIO()
        result.write(self.start)
        if self.separator:
            result.write(self.separator)
        if self.end:
            # Use a default separator if one is not defined.
            if not self.separator:
                result.write("-")
            result.write(self.end)
        return result.getvalue()

    def to_string(self, original: bool = False) -> str:
        """Converts the internal representation of the date range into a
        string. If *original* is :const:`True`, :attr:`original` is returned if
        it exists.

        If :attr:`start` and :attr:`end` are provided without :attr:`separator`,
        ``-`` will be used as default separator in the output.

        This method is identical to calling :func:`str` on this object, provided
        *original* is :const:`False`.
        """
        if original and self.original is not None:
            return self.original
        return str(self)

    def __lt__(self, other: Any) -> bool:
        # pylint: disable=too-many-return-statements
        if not isinstance(other, YearRange):
            return NotImplemented
        # Start year determines most of the storting.
        if self.start < other.start:
            return True
        if self.start > other.start:
            return False
        # Ranges with end dates are sorted after those who don't.
        if self.end is None and other.end is not None:
            return True
        if self.end is not None and other.end is not None:
            # Non-digit ends are sorted after digit ends.
            if not is_four_digits(self.end) and is_four_digits(other.end):
                return False
            if is_four_digits(self.end) and not is_four_digits(other.end):
                return True
            return self.end < other.end
        # No comparing on separators.
        return False

    # TODO: In Python 3.11, use Self
    @classmethod
    def compact(cls, ranges: Iterable["YearRange"]) -> tuple["YearRange", ...]:
        """Given an iterable of :class:`YearRange`, compact them such that a new
        more concise list is returne without losing information. This process
        also sorts the ranges, such that ranges with lower starts come before
        ranges with higher starts.

        - Consecutive years (e.g. 2017, 2018, 2019) are turned into a single
          range (2017-2019).
        - Two consecutive years (e.g. 2017, 2018) are NOT turned turned into a
          single range.
        - Consecutive ranges (e.g. 2017-2019, 2020-2022) are turned into a
          single range (2017-2022).
        - Overlapping ranges (e.g. 2017-2022, 2019-2021) are turned into a
          single range (2017-2022).
        - Repeated ranges are removed.
        - Ranges with non-year ends (e.g. 2017-Present, 2020-Present) are only
          turned into a single range with ranges that have identical ends
          (2017-Present).
        """
        ranges = sorted(ranges)
        compacted: list[YearRange] = []

        if not ranges:
            return tuple()

        # TODO: In Python 3.11, use Self
        def filter_same_end(ranges: Iterable[YearRange]) -> list[YearRange]:
            """If some year ranges have the same end, then only take the ones
            with the lowest start.
            """
            result: list[YearRange] = []
            ends: defaultdict[str | None, list[YearRange]] = defaultdict(list)
            for item in ranges:
                ends[item.end].append(item)
            for key, range_list in ends.items():
                if key is None:
                    for item in range_list:
                        result.append(item)
                else:
                    # *ranges* is sorted, so take the first item with the lowest
                    # start.
                    result.append(range_list[0])
            return sorted(result)

        ranges = filter_same_end(ranges)

        current_start: int = int(ranges[0].start)
        current_end: str = (
            ranges[0].end if ranges[0].end is not None else ranges[0].start
        )

        # TODO: In Python 3.11, use Self
        def add_to_compacted(
            start: int, end: str, next_range: YearRange | None
        ) -> None:
            nonlocal compacted
            if is_four_digits(end):
                if int(end) - start == 0:
                    compacted.append(cls(validate_four_digits(str(start))))
                elif int(end) - start == 1:
                    compacted.append(cls(validate_four_digits(str(start))))
                    compacted.append(cls(validate_four_digits(end)))
                else:
                    compacted.append(
                        cls(
                            validate_four_digits(str(start)),
                            end=end,
                        )
                    )
            else:
                compacted.append(cls(validate_four_digits(str(start)), end=end))
            nonlocal current_start
            nonlocal current_end
            if next_range is not None:
                current_start = int(next_range.start)
                current_end = (
                    next_range.end
                    if next_range.end is not None
                    else next_range.start
                )

        next_start: int | None = None
        next_end: str | None = None
        for next_range in ranges[1:]:
            next_start = int(next_range.start)
            next_end = (
                next_range.end
                if next_range.end is not None
                else next_range.start
            )

            current_end_int: int | None = (
                int(current_end) if is_four_digits(current_end) else None
            )
            next_end_int: int | None = (
                int(next_end) if is_four_digits(next_end) else None
            )
            # The end lines up with next start.
            if (
                current_end_int is not None
                and current_end_int + 1 >= next_start
            ):
                if next_end_int is not None:
                    # In a strange scenario where the next range's end is BEFORE
                    # its start, just save our progress and continue the loop.
                    if next_end_int < next_start:
                        add_to_compacted(current_start, current_end, next_range)
                    # Increment the end.
                    elif next_end_int >= current_end_int:
                        current_end = next_end
            else:
                add_to_compacted(current_start, current_end, next_range)

        add_to_compacted(current_start, current_end, None)
        # If the beforelast range's end was int-y, and the last range's end is
        # string-y, we need to separately add the last range.
        if (
            next_end is not None
            and is_four_digits(current_end)
            and not is_four_digits(next_end)
        ):
            add_to_compacted(cast(int, next_start), next_end, None)
        return tuple(compacted)


def _most_common_prefix(
    copyright_notices: Iterable["CopyrightNotice"],
) -> CopyrightPrefix:
    """Given a number of :class:`CopyrightNotice`s, find the most common one. If
    there is a tie for the most common prefix, return enum which is defined
    before all others in :class:`CopyrightPrefix`.
    """
    counter = Counter(notice.prefix for notice in copyright_notices)
    max_count = max(counter.values())
    most_common = {key for key, value in counter.items() if value == max_count}
    # One prefix is more common than all others.
    if len(most_common) == 1:
        return next(iter(most_common))
    # Enums preserve order of their members. Return the first match.
    for prefix in CopyrightPrefix:
        if prefix in most_common:
            return prefix
    # This shouldn't be reached.
    return CopyrightPrefix.SPDX


@dataclass(frozen=True)
class CopyrightNotice:
    """Represents a single copyright notice."""

    #: The copyright holder. Strictly, this is all text in the copyright notice
    #: which is not part of *years*.
    name: str
    #: The prefix with which the copyright statement begins.
    prefix: CopyrightPrefix = CopyrightPrefix.SPDX
    #: The dates associated with the copyright notice.
    years: tuple[YearRange, ...] = field(default_factory=tuple)

    #: If parsed from a string, this contains the original string.
    original: str | None = field(
        default=None, init=False, repr=False, compare=False
    )

    # TODO: In Python 3.11, return Self.
    @classmethod
    def from_string(cls, value: str) -> "CopyrightNotice":
        """Create a :class:`CopyrightNotice` object from a string.

        Raises:
            CopyrightNoticeParseError: The string is not a valid copyright
                notice.
        """
        re_result = COPYRIGHT_NOTICE_PATTERN.fullmatch(value)
        if not re_result:
            raise CopyrightNoticeParseError(
                f"'{value}' is not a copyright notice."
            )
        return cls.from_match(re_result)

    @staticmethod
    def _detect_prefix(prefix: str) -> CopyrightPrefix:
        """Given a matched prefix from :const:`COPYRIGHT_NOTICE_PATTERN`, detect
        the associated prefix.
        """
        prefix_lower = prefix.lower()
        # String-match the prefix.
        for prefix_enum in (
            cast(CopyrightPrefix, item) for item in reversed(CopyrightPrefix)
        ):
            # lower() is used to match (C) as well as (c).
            if prefix_lower == prefix_enum.value.lower():
                return prefix_enum
        # The prefix could not be string-matched, most likely because there
        # was unexpected spacing in the prefix. Get a close match using
        # difflib.
        matches = difflib.get_close_matches(
            prefix,
            # TODO: In Python 3.11, this list comprehension is not needed.
            [item.value for item in CopyrightPrefix],
            n=1,
            cutoff=0.2,
        )
        if matches:
            return CopyrightPrefix(matches[0])
        # This shouldn't happen, but if no prefix could be found,
        # default to SPDX.
        return CopyrightPrefix.SPDX

    @classmethod
    def from_match(cls, value: re.Match) -> "CopyrightNotice":
        """Create a :class:`CopyrightNotice` object from a regular expression
        match using the :const:`COPYRIGHT_NOTICE_PATTERN` :class:`re.Pattern`.
        """
        prefix = cls._detect_prefix(value.group("prefix"))

        re_text = value.group("text")
        year_ranges_substrings = list(_YEARS_PATTERN.finditer(re_text))
        start_ends: list[tuple[int, int]] = [
            (match.start("prefix"), match.end("years"))
            for match in year_ranges_substrings
        ]
        name_parts: list[str] = []
        last = 0
        for start, end in start_ends:
            name_parts.append(re_text[last:start])
            last = end
        name_parts.append(re_text[last:])

        name = "".join(name_parts)
        # Remove ', ' from the start of the name, which appears in e.g.
        # 'Copyright 2017, Jane Doe'.
        name = _COMMA_SPACE_PATTERN.sub("", name)
        years: tuple[YearRange, ...] = tuple()
        if year_ranges_substrings:
            years = tuple(
                chain.from_iterable(
                    YearRange.tuple_from_string(match.group("years"))
                    for match in year_ranges_substrings
                )
            )
        result = cls(
            name=name,
            prefix=prefix,
            years=years,
        )
        object.__setattr__(result, "original", value.string)
        return result

    @classmethod
    def merge(
        cls, copyright_notices: Iterable["CopyrightNotice"]
    ) -> set["CopyrightNotice"]:
        """Given an iterable of :class:`CopyrightNotice`, merge all notices
        which have the same name. The years are compacted, and from the
        :class:`CopyrightPrefix` prefixes in *copyright_notices*, the most
        common is chosen. If there is a tie in frequency, choose the one which
        appears first in the enum.
        """
        matches: defaultdict[str, list[CopyrightNotice]] = defaultdict(list)
        result: set[CopyrightNotice] = set()
        for notice in copyright_notices:
            matches[notice.name].append(notice)
        for key, value in matches.items():
            result.add(
                cls(
                    key,
                    prefix=_most_common_prefix(value),
                    years=YearRange.compact(
                        chain.from_iterable(notice.years for notice in value)
                    ),
                )
            )
        return result

    def __str__(self) -> str:
        result = StringIO()
        result.write(self.prefix.value)
        if self.years:
            result.write(" ")
            result.write(
                ", ".join(str(date_range) for date_range in self.years)
            )
        result.write(f" {self.name}")
        return result.getvalue()

    def __lt__(self, other: "CopyrightNotice") -> bool:
        def norm(item: Any | None) -> tuple[int, Any]:
            """If no item is defined, return a tuple that sorts _after_
            items that are defined.
            """
            return (0, item) if item else (1, "")

        return (
            norm(self.years),
            self.name,
            self.prefix.value,
        ) < (
            norm(other.years),
            other.name,
            other.prefix.value,
        )

    def to_string(self, original: bool = False) -> str:
        """Converts the internal representation of the copyright notice into a
        string. If *original* is :const:`True`, :attr:`original` is returned if
        it exists.

        This method is identical to calling :func:`str` on this object, provided
        *original* is :const:`False`.
        """
        if original and self.original is not None:
            return self.original
        return str(self)


@dataclass(frozen=True)
class SpdxExpression:
    """A simple dataclass that contains an SPDX License Expression.

    Use :meth:`SpdxExpression.__str__` to get a string representation of the
    expression.
    """

    #: A string representing an SPDX License Expression. It may be invalid.
    text: InitVar[str]
    _text: str = field(init=False, repr=True)

    def __post_init__(self, text: str) -> None:
        object.__setattr__(self, "_text", text)

    @cached_property
    def is_valid(self) -> bool:
        """If :attr:`text` is a valid SPDX License Expression, this property is
        :const:`True`.

        To be 'valid', it has to follow the grammar and syntax of the SPDX
        specification. The licenses and exceptions need not appear on the
        license list.
        """
        return self._expression is not None

    @cached_property
    def _expression(self) -> LicenseExpression | None:
        """A parsed :class:`LicenseExpression` from :attr:`text`. If
        :attr:`text` could not be parsed, *_expression*'s value is
        :const:`None`.
        """
        try:
            return _LICENSING.parse(self._text, simple=True)
        except ExpressionError:
            return None

    @cached_property
    def licenses(self) -> list[str]:
        """Return a list of licenses used in the expression, in order of
        appearance, without duplicates.

        If the expression is invalid, the list contains a single item
        :attr:`text`.
        """
        if self._expression is not None:
            return _LICENSING.license_keys(self._expression)
        return [self._text]

    @classmethod
    def combine(
        cls,
        spdx_expressions: Iterable["SpdxExpression"],
    ) -> "SpdxExpression":
        """Combine the *spdx_expressions* into a single :class:`SpdxExpression`,
        joined by AND operators.
        """
        is_valid = True
        for expression in spdx_expressions:
            if not expression.is_valid:
                is_valid = False
        if is_valid:
            return cls(
                str(
                    combine_expressions(
                        list(
                            # pylint: disable=protected-access
                            expression._expression
                            for expression in spdx_expressions
                        )
                    )
                )
            )
        return cls(
            " AND ".join(f"({expression})" for expression in spdx_expressions)
        )

    def simplify(self) -> "SpdxExpression":
        """If the expression is valid, return a new :class:`SpdxExpression`
        which is 'simplified', meaning that boolean operators are collapsed.
        'MIT OR MIT' simplifies to 'MIT', and so forth.

        If the expression is not valid, ``self`` is returned.
        """
        if self.is_valid:
            return self.__class__(
                str(cast(LicenseExpression, self._expression).simplify())
            )
        return self

    def __str__(self) -> str:
        """Return a string representation of the expression if it is valid.
        Otherwise, return :attr:`text`.
        """
        if self._expression is not None:
            return str(self._expression)
        return self._text

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, SpdxExpression):
            return NotImplemented
        if self._expression is not None and other._expression is not None:
            return self._expression == other._expression
        return self._text == other._text

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, SpdxExpression):
            return NotImplemented
        return str(self) < str(other)


class SourceType(Enum):
    """
    An enumeration representing the types of sources for license information.
    """

    #: A .license file containing license information.
    DOT_LICENSE = "dot-license"
    #: A file header containing license information.
    FILE_HEADER = "file-header"
    #: A .reuse/dep5 file containing license information.
    DEP5 = "dep5"
    #: A REUSE.toml file containing license information.
    REUSE_TOML = "reuse-toml"


@dataclass(frozen=True, kw_only=True)
class ReuseInfo:
    """Simple dataclass holding licensing and copyright information"""

    spdx_expressions: set[SpdxExpression] = field(default_factory=set)
    copyright_notices: set[CopyrightNotice] = field(default_factory=set)
    contributor_lines: set[str] = field(default_factory=set)
    path: str | None = None
    source_path: str | None = None
    source_type: SourceType | None = None

    def _check_nonexistent(self, **kwargs: Any) -> None:
        nonexistent_attributes = set(kwargs) - set(self.__dict__)
        if nonexistent_attributes:
            raise KeyError(
                f"The following attributes do not exist in"
                f" {self.__class__}: {', '.join(nonexistent_attributes)}"
            )

    def copy(self, **kwargs: Any) -> "ReuseInfo":
        """Return a copy of ReuseInfo, replacing the values of attributes with
        the values from *kwargs*.
        """
        self._check_nonexistent(**kwargs)
        new_kwargs = {}
        for key, value in self.__dict__.items():
            new_kwargs[key] = kwargs.get(key, value)
        return self.__class__(**new_kwargs)  # type: ignore

    def union(self, *other: "ReuseInfo") -> "ReuseInfo":
        """Return a new instance of ReuseInfo where all set attributes are equal
        to the union of the set in *self* and the set(s) in *other*.

        All non-set attributes are set to their values in *self*.

        >>> one = ReuseInfo(copyright_notices={CopyrightNotice("Jane Doe")},
        ...           source_path="foo.py")
        >>> two = ReuseInfo(copyright_notices={CopyrightNotice("John Doe")},
        ...           source_path="bar.py")
        >>> result = one.union(two)
        >>> print([notice.name for notice in sorted(result.copyright_notices)])
        ['Jane Doe', 'John Doe']
        >>> print(result.source_path)
        foo.py
        """
        if not other:
            return self
        new_kwargs = {}
        for key, attr_val in self.__dict__.items():
            if isinstance(attr_val, set):
                new_kwargs[key] = attr_val.union(
                    *(getattr(info, key) for info in other)
                )
            else:
                new_kwargs[key] = attr_val
        return self.__class__(**new_kwargs)  # type: ignore

    def contains_copyright_or_licensing(self) -> bool:
        """Either *spdx_expressions* or *copyright_notices* is non-empty."""
        return bool(self.spdx_expressions or self.copyright_notices)

    def contains_copyright_xor_licensing(self) -> bool:
        """One of *spdx_expressions* or *copyright_notices* is non-empty."""
        return bool(self.spdx_expressions) ^ bool(self.copyright_notices)

    def contains_info(self) -> bool:
        """Any field except *path*, *source_path* and *source_type* is
        non-empty.
        """
        keys = {
            key
            for key in self.__dict__
            if key not in ("path", "source_path", "source_type")
        }
        return any(self.__dict__[key] for key in keys)

    def __bool__(self) -> bool:
        return any(self.__dict__.values())

    def __or__(self, value: "ReuseInfo") -> "ReuseInfo":
        return self.union(value)


# REUSE-IgnoreEnd



# SPDX-FileCopyrightText: 2019 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2019 Stefan Bakker <s.bakker777@gmail.com>
# SPDX-FileCopyrightText: 2019 Kirill Elagin <kirelagin@gmail.com>
# SPDX-FileCopyrightText: 2020 Dmitry Bogatov
# SPDX-FileCopyrightText: © 2020 Liferay, Inc. <https://liferay.com>
# SPDX-FileCopyrightText: 2021 Alvar Penning
# SPDX-FileCopyrightText: 2021 Alliander N.V. <https://alliander.com>
# SPDX-FileCopyrightText: 2021 Robin Vobruba <hoijui.quaero@gmail.com>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2022 Yaman Qalieh
# SPDX-FileCopyrightText: 2022 Carmen Bianca Bakker <carmenbianca@fsfe.org>
# SPDX-FileCopyrightText: 2025 Rivos Inc.
# SPDX-FileCopyrightText: 2025 Matthias Schoettle <opensource@mattsch.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Functions for manipulating the comment headers of files."""

import logging
import re
from collections.abc import Sequence
from typing import NamedTuple, cast

from jinja2 import Environment, PackageLoader, Template

from .comment import (
    CommentStyle,
    EmptyCommentStyle,
    HtmlCommentStyle,
    PythonCommentStyle,
)
from .copyright import CopyrightNotice, ReuseInfo
from .exceptions import CommentParseError, MissingReuseInfoError
from .extract import contains_reuse_info, extract_reuse_info
from .i18n import _

_LOGGER = logging.getLogger(__name__)

_ENV = Environment(loader=PackageLoader("reuse", "templates"), trim_blocks=True)
DEFAULT_TEMPLATE = _ENV.get_template("default_template.jinja2")

_NEWLINE_PATTERN = re.compile(r"\n", re.MULTILINE)


class _TextSections(NamedTuple):
    """Used to split up text in three parts."""

    before: str
    middle: str
    after: str


class _FrontmatterCommentStyle(CommentStyle):
    """Frontmatter comment style. Used specifically for e.g. Markdown."""

    SHORTHAND = "frontmatter"

    SINGLE_LINE = "#"
    INDENT_AFTER_SINGLE = " "
    SHEBANGS = ["---"]


def _create_new_header(
    reuse_info: ReuseInfo,
    template: Template | None = None,
    template_is_commented: bool = False,
    style: type[CommentStyle] | None = None,
    force_multi: bool = False,
) -> str:
    """Format a new header from scratch.

    Raises:
        CommentCreateError: if a comment could not be created.
        MissingReuseInfoError: if the generated comment is missing SPDX
            information.
    """
    if template is None:
        template = DEFAULT_TEMPLATE
    if style is None:
        style = cast(type[CommentStyle], PythonCommentStyle)

    rendered = template.render(
        copyright_lines=map(str, sorted(reuse_info.copyright_notices)),
        contributor_lines=sorted(reuse_info.contributor_lines),
        spdx_expressions=sorted(map(str, reuse_info.spdx_expressions)),
    ).strip("\n")

    if template_is_commented:
        result = rendered
    else:
        result = style.create_comment(rendered, force_multi=force_multi).strip(
            "\n"
        )

    # Verify that the result contains all ReuseInfo.
    new_reuse_info = extract_reuse_info(result)
    if (
        reuse_info.copyright_notices != new_reuse_info.copyright_notices
        and reuse_info.spdx_expressions != new_reuse_info.spdx_expressions
    ):
        _LOGGER.debug(
            _(
                "generated comment is missing copyright lines or license"
                " expressions"
            )
        )
        _LOGGER.debug(result)
        raise MissingReuseInfoError()

    return result


# pylint: disable=too-many-arguments
def create_header(
    reuse_info: ReuseInfo,
    header: str | None = None,
    template: Template | None = None,
    template_is_commented: bool = False,
    style: type[CommentStyle] | None = None,
    force_multi: bool = False,
    merge_copyrights: bool = False,
) -> str:
    """Create a header containing *reuse_info*. *header* is an optional argument
    containing a header which should be modified to include *reuse_info*. If
    *header* is not given, a brand new header is created.

    *template*, *template_is_commented*, and *style* determine what the header
    will look like, and whether it will be commented or not.

    Raises:
        CommentCreateError: if a comment could not be created.
        MissingReuseInfoError: if the generated comment is missing SPDX
            information.
    """
    if template is None:
        template = DEFAULT_TEMPLATE
    if style is None:
        style = PythonCommentStyle

    new_header = ""
    if header:
        existing_spdx = extract_reuse_info(header)
        if merge_copyrights:
            spdx_copyrights = CopyrightNotice.merge(
                reuse_info.copyright_notices.union(
                    existing_spdx.copyright_notices
                )
            )
        else:
            spdx_copyrights = reuse_info.copyright_notices.union(
                existing_spdx.copyright_notices
            )

        # TODO: This behaviour does not match the docstring.
        reuse_info = existing_spdx | reuse_info
        reuse_info = reuse_info.copy(copyright_notices=spdx_copyrights)

    new_header += _create_new_header(
        reuse_info,
        template=template,
        template_is_commented=template_is_commented,
        style=style,
        force_multi=force_multi,
    )
    return new_header


def _indices_of_newlines(text: str) -> Sequence[int]:
    indices = [0]
    start = 0

    while True:
        match = _NEWLINE_PATTERN.search(text, start)
        if match:
            start = match.span()[1]
            indices.append(start)
        else:
            break

    return indices


def _find_first_spdx_comment(
    text: str, style: type[CommentStyle] | None = None
) -> _TextSections:
    """Find the first SPDX comment in the file. Return a tuple with everything
    preceding the comment, the comment itself, and everything following it.

    Raises:
        MissingReuseInfoError: if no REUSE info can be found in any comment.
    """
    if style is None:
        style = PythonCommentStyle

    indices = _indices_of_newlines(text)

    for index in indices:
        try:
            comment = style.comment_at_first_character(text[index:])
        except CommentParseError:
            continue
        if contains_reuse_info(comment):
            return _TextSections(
                text[:index], comment + "\n", text[index + len(comment) + 1 :]
            )

    raise MissingReuseInfoError()


def _extract_shebang(prefix: str, text: str) -> tuple[str, str]:
    """Remove all lines that start with the shebang prefix from *text*. Return a
    tuple of (shebang, reduced_text).
    """
    shebang_lines = []
    for line in text.splitlines(keepends=True):
        if line.startswith(prefix):
            shebang_lines.append(line)
            text = text.replace(line, "", 1)
        else:
            break
    shebang = "".join(shebang_lines)
    return (shebang, text)


def place_header(
    header: str,
    before: str,
    after: str,
    has_existing_header: bool,
    double_newline_after_before: bool = True,
) -> str:
    """Construct the resulting file with the header and the rest of the text
    in the file.
    """
    new_text = f"{header}\n"
    if before.strip():
        newlines = "\n\n" if double_newline_after_before else "\n"
        new_text = f"{before.rstrip()}{newlines}{new_text}"
    if after.strip():
        # Create space between header and following code only if a newline
        # doesn't already exist, and there wasn't previously a header.
        if not has_existing_header and not after.startswith("\n"):
            separator = "\n"
        else:
            separator = ""

        new_text = f"{new_text}{separator}{after}"
    return new_text


# pylint: disable=too-many-arguments
def find_and_replace_header(
    text: str,
    reuse_info: ReuseInfo,
    template: Template | None = None,
    template_is_commented: bool = False,
    style: type[CommentStyle] | None = None,
    force_multi: bool = False,
    merge_copyrights: bool = False,
) -> str:
    """Find the first SPDX comment block in *text*. That comment block is
    replaced by a new comment block containing *reuse_info*. It is formatted as
    according to *template*. The template is normally uncommented, but if it is
    already commented, *template_is_commented* should be :const:`True`.

    If both *style* and *template_is_commented* are provided, *style* is only
    used to find the header comment.

    If the comment block already contained some REUSE information, that
    information is merged into *reuse_info*.

    If no header exists, one is simply created.

    *text* is returned with a new header.

    Raises:
        CommentCreateError: if a comment could not be created.
        MissingReuseInfoError: if the generated comment is missing SPDX
            information.
    """
    if style is None:
        style = PythonCommentStyle

    # Workaround: Treat Markdown files with frontmatter with the Frontmatter
    # style instead.
    frontmatter = text.startswith("---")
    if style is HtmlCommentStyle and frontmatter:
        style = _FrontmatterCommentStyle

    try:
        before, header, after = _find_first_spdx_comment(text, style=style)
    except MissingReuseInfoError:
        before, header, after = "", "", text

    # Workaround. EmptyCommentStyle should always be completely replaced.
    if style is EmptyCommentStyle:
        after = ""

    _LOGGER.debug(f"before = {repr(before)}")
    _LOGGER.debug(f"header = {repr(header)}")
    _LOGGER.debug(f"after = {repr(after)}")

    # Keep special first-line-of-file lines as the first line in the file,
    # or say, move our comments after it.
    if style.SHEBANGS:
        for shebang in style.SHEBANGS:
            # Extract shebang from header and put it in before. It's a bit
            # messy, but it ends up working.
            if header.startswith(shebang) and not before.strip():
                before, header = _extract_shebang(shebang, header)
            elif after.startswith(shebang) and not any((before, header)):
                before, after = _extract_shebang(shebang, after)
            else:
                continue
            break

    new_header = create_header(
        reuse_info,
        header,
        template=template,
        template_is_commented=template_is_commented,
        style=style,
        force_multi=force_multi,
        merge_copyrights=merge_copyrights,
    )

    return place_header(
        new_header,
        before,
        after,
        bool(header),
        double_newline_after_before=not frontmatter,
    )


# pylint: disable=too-many-arguments
def add_new_header(
    text: str,
    reuse_info: ReuseInfo,
    template: Template | None = None,
    template_is_commented: bool = False,
    style: type[CommentStyle] | None = None,
    force_multi: bool = False,
    merge_copyrights: bool = False,
) -> str:
    """Add a new header at the very top of *text*, similar to
    find_and_replace_header. But in this function, do not replace any headers or
    search for any existing REUSE information.

    Raises:
        CommentCreateError: if a comment could not be created.
    """
    if style is None:
        style = PythonCommentStyle

    shebang = ""

    if style.SHEBANGS:
        for shebang_prefix in style.SHEBANGS:
            if text.startswith(shebang_prefix):
                shebang, text = _extract_shebang(shebang_prefix, text)
                break

    header = create_header(
        reuse_info,
        None,
        template=template,
        template_is_commented=template_is_commented,
        style=style,
        force_multi=force_multi,
        merge_copyrights=merge_copyrights,
    )

    return place_header(header, shebang, text, False)



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""The REUSE Specification has a concept called Covered Files; files which must
contain licensing information. Some files in a project are not Covered Files,
and thus needn't contain licensing information. This module contains all that
logic.
"""

import contextlib
import logging
import os
import re
import stat
from collections.abc import Collection, Generator
from pathlib import Path
from typing import cast

from .types import StrPath
from .vcs import VCSStrategy

_LOGGER = logging.getLogger(__name__)

_IGNORE_DIR_PATTERNS = [
    re.compile(r"^\.git$"),
    re.compile(r"^\.hg$"),
    re.compile(r"^\.sl$"),  # Used by Sapling SCM
    re.compile(r"^LICENSES$"),
    re.compile(r"^\.reuse$"),
]

_IGNORE_MESON_PARENT_DIR_PATTERNS = [
    re.compile(r"^subprojects$"),
]

_IGNORE_FILE_PATTERNS = [
    # LICENSE, LICENSE-MIT, LICENSE.txt
    re.compile(r"^LICEN[CS]E([-\.].*)?$"),
    re.compile(r"^COPYING([-\.].*)?$"),
    # ".git" as file happens in submodules
    re.compile(r"^\.git$"),
    re.compile(r"^\.hgtags$"),
    re.compile(r".*\.license$"),
    re.compile(r"^REUSE\.toml$"),
    # Workaround for https://github.com/fsfe/reuse-tool/issues/229
    re.compile(r"^CAL-1.0(-Combined-Work-Exception)?(\..+)?$"),
    re.compile(r"^SHL-2.1(\..+)?$"),
]

_IGNORE_SPDX_PATTERNS = [
    # SPDX files from
    # https://spdx.github.io/spdx-spec/conformance/#44-standard-data-format-requirements
    re.compile(r".*\.spdx$"),
    re.compile(r".*\.spdx.(rdf|json|xml|ya?ml)$"),
]

# Combine SPDX patterns into file patterns to ease default ignore usage
_IGNORE_FILE_PATTERNS.extend(_IGNORE_SPDX_PATTERNS)


def is_path_ignored(
    path: Path,
    subset_files: Collection[StrPath] | None = None,
    include_submodules: bool = False,
    include_meson_subprojects: bool = False,
    include_reuse_tomls: bool = False,
    vcs_strategy: VCSStrategy | None = None,
) -> bool:
    """Is *path* ignored by some mechanism?"""
    # pylint: disable=too-many-return-statements,too-many-branches
    name = path.name

    # Only stat the file once instead of multiple times.
    stat_result = path.lstat()

    # Symlink.
    if stat.S_ISLNK(stat_result.st_mode):
        _LOGGER.debug("skipping symlink '%s'", path)
        return True
    # File.
    if stat.S_ISREG(stat_result.st_mode):
        if subset_files is not None and path.resolve() not in subset_files:
            return True
        for pattern in _IGNORE_FILE_PATTERNS:
            if pattern.match(name) and (
                name != "REUSE.toml" or not include_reuse_tomls
            ):
                return True
        # Suppressing this error because I simply don't want to deal
        # with that here.
        with contextlib.suppress(OSError):
            if stat_result.st_size == 0:
                _LOGGER.debug("skipping 0-sized file '%s'", path)
                return True
    # Directory.
    elif stat.S_ISDIR(stat_result.st_mode):
        if subset_files is not None and not any(
            Path(file_).is_relative_to(path.resolve()) for file_ in subset_files
        ):
            return True
        for pattern in _IGNORE_DIR_PATTERNS:
            if pattern.match(name):
                return True
        if not include_meson_subprojects:
            for pattern in _IGNORE_MESON_PARENT_DIR_PATTERNS:
                if pattern.match(path.parent.name):
                    _LOGGER.info(
                        "ignoring '%s' because it is a Meson subproject", path
                    )
                    return True
        if (
            not include_submodules
            and vcs_strategy
            and vcs_strategy.is_submodule(path)
        ):
            _LOGGER.info("ignoring '%s' because it is a submodule", path)
            return True

    if vcs_strategy and vcs_strategy.is_ignored(path):
        return True

    return False


def iter_files(
    directory: StrPath,
    subset_files: Collection[StrPath] | None = None,
    include_submodules: bool = False,
    include_meson_subprojects: bool = False,
    include_reuse_tomls: bool = False,
    vcs_strategy: VCSStrategy | None = None,
) -> Generator[Path, None, None]:
    """Yield all Covered Files in *directory* and its subdirectories according
    to the REUSE Specification.
    """
    directory = Path(directory)
    if subset_files is not None:
        subset_files = cast(
            set[Path], {Path(file_).resolve() for file_ in subset_files}
        )

    for root_str, dirs, files in os.walk(directory):
        root = Path(root_str)

        # Don't walk ignored directories
        for dir_ in list(dirs):
            the_dir = root / dir_
            if is_path_ignored(
                the_dir,
                subset_files=subset_files,
                include_submodules=include_submodules,
                include_meson_subprojects=include_meson_subprojects,
                include_reuse_tomls=include_reuse_tomls,
                vcs_strategy=vcs_strategy,
            ):
                _LOGGER.debug("ignoring '%s'", the_dir)
                dirs.remove(dir_)

        # Filter files.
        for file_ in files:
            the_file = root / file_
            if is_path_ignored(
                the_file,
                subset_files=subset_files,
                include_submodules=include_submodules,
                include_meson_subprojects=include_meson_subprojects,
                include_reuse_tomls=include_reuse_tomls,
                vcs_strategy=vcs_strategy,
            ):
                _LOGGER.debug("ignoring '%s'", the_file)
                continue

            yield the_file



# SPDX-FileCopyrightText: 2019 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2019 Kirill Elagin <kirelagin@gmail.com>
# SPDX-FileCopyrightText: 2019 Stefan Bakker <s.bakker777@gmail.com>
# SPDX-FileCopyrightText: 2020 Dmitry Bogatov
# SPDX-FileCopyrightText: 2021 Alliander N.V. <https://alliander.com>
# SPDX-FileCopyrightText: 2021 Alvar Penning
# SPDX-FileCopyrightText: 2021 Robin Vobruba <hoijui.quaero@gmail.com>
# SPDX-FileCopyrightText: 2022 Carmen Bianca Bakker <carmenbianca@fsfe.org>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2022 Yaman Qalieh
# SPDX-FileCopyrightText: 2024 Rivos Inc.
# SPDX-FileCopyrightText: © 2020 Liferay, Inc. <https://liferay.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Functions for the CLI portion of manipulating headers."""

import logging
import sys
from typing import IO, cast

from jinja2 import Environment, FileSystemLoader, Template
from jinja2.exceptions import TemplateNotFound

from ._util import _determine_license_suffix_path
from .comment import (
    NAME_STYLE_MAP,
    CommentStyle,
    EmptyCommentStyle,
    get_comment_style,
)
from .copyright import ReuseInfo
from .exceptions import CommentCreateError, MissingReuseInfoError
from .extract import contains_reuse_info
from .header import add_new_header, find_and_replace_header
from .i18n import _
from .project import Project
from .types import StrPath

_LOGGER = logging.getLogger(__name__)


def find_template(project: Project, name: str) -> Template:
    """Find a template given a name.

    Raises:
        TemplateNotFound: if template could not be found.
    """
    template_dir = project.root / ".reuse/templates"
    env = Environment(
        loader=FileSystemLoader(str(template_dir)), trim_blocks=True
    )

    names = [name]
    if not name.endswith(".jinja2"):
        names.append(f"{name}.jinja2")
    if not name.endswith(".commented.jinja2"):
        names.append(f"{name}.commented.jinja2")

    for item in names:
        try:
            return env.get_template(item)
        except TemplateNotFound:
            pass
    raise TemplateNotFound(name)


def add_header_to_file(
    path: StrPath,
    reuse_info: ReuseInfo,
    template: Template | None,
    template_is_commented: bool,
    style: str | None,
    encoding: str = "utf-8",
    newline: str = "\n",
    force_multi: bool = False,
    skip_existing: bool = False,
    skip_unrecognised: bool = False,
    fallback_dot_license: bool = False,
    merge_copyrights: bool = False,
    replace: bool = True,
    out: IO[str] = sys.stdout,
) -> int:
    """Helper function."""
    # pylint: disable=too-many-arguments,too-many-locals
    result = 0
    comment_style: type[CommentStyle] | None = NAME_STYLE_MAP.get(
        cast(str, style)
    )
    if comment_style is None:
        comment_style = get_comment_style(path)
    if comment_style is None:
        if skip_unrecognised:
            out.write(_("Skipped unrecognised file '{path}'").format(path=path))
            out.write("\n")
            return result
        if fallback_dot_license:
            out.write(
                _(
                    "'{path}' is not recognised; creating '{path}.license'"
                ).format(path=path)
            )
            out.write("\n")
            path = _determine_license_suffix_path(path)
            path.touch()
            comment_style = EmptyCommentStyle

    with open(path, encoding=encoding) as fp:
        text = fp.read()

    # Ideally, this check is done elsewhere. But that would necessitate reading
    # the file contents before this function is called.
    if skip_existing and contains_reuse_info(text):
        out.write(
            _(
                "Skipped file '{path}' already containing REUSE information"
            ).format(path=path)
        )
        out.write("\n")
        return result

    try:
        if replace:
            output = find_and_replace_header(
                text,
                reuse_info,
                template=template,
                template_is_commented=template_is_commented,
                style=comment_style,
                force_multi=force_multi,
                merge_copyrights=merge_copyrights,
            )
        else:
            output = add_new_header(
                text,
                reuse_info,
                template=template,
                template_is_commented=template_is_commented,
                style=comment_style,
                force_multi=force_multi,
                merge_copyrights=merge_copyrights,
            )
    except CommentCreateError:
        out.write(
            _("Error: Could not create comment for '{path}'").format(path=path)
        )
        out.write("\n")
        result = 1
    except MissingReuseInfoError:
        out.write(
            _(
                "Error: Generated comment header for '{path}' is missing"
                " copyright lines or license expressions. The template is"
                " probably incorrect. Did not write new header."
            ).format(path=path)
        )
        out.write("\n")
        result = 1
    else:
        with open(path, "w", encoding=encoding, newline=newline) as fp:
            fp.write(output)
        # TODO: This may need to be rephrased more elegantly.
        out.write(_("Successfully changed header of {path}").format(path=path))
        out.write("\n")

    return result



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2020 Tuomas Siipola <tuomas@zpl.fi>
# SPDX-FileCopyrightText: 2022 Carmen Bianca Bakker <carmenbianca@fsfe.org>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2022 Nico Rikken <nico.rikken@fsfe.org>
# SPDX-FileCopyrightText: 2022 Pietro Albini <pietro.albini@ferrous-systems.com>
# SPDX-FileCopyrightText: 2023 DB Systel GmbH
# SPDX-FileCopyrightText: 2023 Johannes Zarl-Zierl <johannes@zarl-zierl.at>
# SPDX-FileCopyrightText: 2024 Rivos Inc.
# SPDX-FileCopyrightText: 2024 Skyler Grey <sky@a.starrysky.fyi>
# SPDX-FileCopyrightText: © 2020 Liferay, Inc. <https://liferay.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Misc. utilities for reuse."""

import logging
import os
import subprocess
from hashlib import sha1
from inspect import cleandoc
from pathlib import Path
from typing import IO, Any

from .types import StrPath

# REUSE-IgnoreStart


def setup_logging(level: int = logging.WARNING) -> None:
    """Configure logging for reuse.

    You can only call this function once.
    """
    # library_logger is the root logger for reuse. We configure logging solely
    # for reuse, not for any other libraries.
    library_logger = logging.getLogger("reuse")

    if not library_logger.hasHandlers():
        library_logger.setLevel(level)
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
        handler.setFormatter(formatter)
        library_logger.addHandler(handler)


def execute_command(
    command: list[str],
    logger: logging.Logger,
    cwd: StrPath | None = None,
    **kwargs: Any,
) -> subprocess.CompletedProcess:
    """Run the given command with subprocess.run. Forward kwargs. Silence
    output into a pipe unless kwargs override it.
    """
    logger.debug("running '%s'", " ".join(command))

    stdout: None | int | IO[Any] = kwargs.get("stdout", subprocess.PIPE)
    stderr: None | int | IO[Any] = kwargs.get("stderr", subprocess.PIPE)

    return subprocess.run(
        list(map(str, command)),
        stdout=stdout,
        stderr=stderr,
        check=False,
        cwd=str(cwd),
        **kwargs,
    )


def find_licenses_directory(root: StrPath | None = None) -> Path:
    """Find the licenses directory from CWD or *root*. In the following order:

    - LICENSES/ in *root*.

    - Current directory if its name is "LICENSES"

    - LICENSES/ in CWD.

    The returned LICENSES/ directory NEED NOT EXIST. It may still need to be
    created.
    """
    cwd = Path.cwd()
    licenses_path = cwd / "LICENSES"

    if root:
        licenses_path = Path(root) / "LICENSES"
    elif cwd.name == "LICENSES":
        licenses_path = cwd

    return licenses_path


def _determine_license_path(path: StrPath) -> Path:
    """Given a path FILE, return FILE.license if it exists, otherwise return
    FILE.
    """
    license_path = Path(f"{path}.license")
    if not license_path.exists():
        license_path = Path(path)
    return license_path


def _determine_license_suffix_path(path: StrPath) -> Path:
    """Given a path FILE or FILE.license, return FILE.license."""
    path = Path(path)
    if path.suffix == ".license":
        return path
    return Path(f"{path}.license")


def _strip_plus_from_identifier(spdx_identifier: str) -> str:
    """Strip final plus from identifier.

    >>> _strip_plus_from_identifier("EUPL-1.2+")
    'EUPL-1.2'
    >>> _strip_plus_from_identifier("EUPL-1.2")
    'EUPL-1.2'
    """
    if spdx_identifier.endswith("+"):
        return spdx_identifier[:-1]
    return spdx_identifier


def _add_plus_to_identifier(spdx_identifier: str) -> str:
    """Add final plus to identifier.

    >>> _add_plus_to_identifier("EUPL-1.2")
    'EUPL-1.2+'
    >>> _add_plus_to_identifier("EUPL-1.2+")
    'EUPL-1.2+'
    """
    if spdx_identifier.endswith("+"):
        return spdx_identifier
    return f"{spdx_identifier}+"


def relative_from_root(path: Path, root: Path) -> Path:
    """A helper function to get *path* relative to *root*."""
    path_parts = path.parts
    root_parts = root.parts
    root_parts_len = len(root_parts)

    # This is rather strangely more performant than `path.relative_to(root)`.
    if path_parts[:root_parts_len] == root_parts:
        return Path(*path_parts[root_parts_len:])
    return Path(os.path.relpath(str(path), start=str(root)))


def _checksum(path: StrPath) -> str:
    path = Path(path)

    file_sha1 = sha1()
    with path.open("rb") as fp:
        for chunk in iter(lambda: fp.read(128 * file_sha1.block_size), b""):
            file_sha1.update(chunk)

    return file_sha1.hexdigest()


def cleandoc_nl(text: str) -> str:
    """Like :func:`inspect.cleandoc`, but with a newline at the end."""
    return cleandoc(text) + "\n"


# REUSE-IgnoreEnd



# SPDX-FileCopyrightText: 2024 Carmen Bianca BAKKER <carmenbianca@fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Logic to convert a .reuse/dep5 file to a REUSE.toml file."""

import re
from collections.abc import Iterable
from typing import Any, Optional, TypeVar, cast

import tomlkit
from debian.copyright import Copyright, FilesParagraph, Header

from .global_licensing import REUSE_TOML_VERSION

_SINGLE_ASTERISK_PATTERN = re.compile(r"(?<!\*)\*(?!\*)")

_T = TypeVar("_T")


def _collapse_list_if_one_item(
    # Technically this should be Sequence[_T], but I can't get that to work.
    sequence: list[_T],
) -> list[_T] | _T:
    """Return the only item of the list if the length of the list is one, else
    return the list.
    """
    if len(sequence) == 1:
        return sequence[0]
    return sequence


def _header_from_dep5_header(
    header: Header,
) -> dict[str, str | list[str]]:
    result: dict[str, str | list[str]] = {}
    if header.upstream_name:
        result["SPDX-PackageName"] = str(header.upstream_name)
    if header.upstream_contact:
        result["SPDX-PackageSupplier"] = _collapse_list_if_one_item(
            list(map(str, header.upstream_contact))
        )
    if header.source:
        result["SPDX-PackageDownloadLocation"] = str(header.source)
    if header.disclaimer:
        result["SPDX-PackageComment"] = str(header.disclaimer)
    return result


def _copyrights_from_paragraph(
    paragraph: FilesParagraph,
) -> str | list[str]:
    return _collapse_list_if_one_item(
        [line.strip() for line in cast(str, paragraph.copyright).splitlines()]
    )


def _convert_asterisk(path: str) -> str:
    """This solves a semantics difference. A singular asterisk is semantically
    identical to a double asterisk in REUSE.toml.
    """
    return _SINGLE_ASTERISK_PATTERN.sub("**", path)


def _paths_from_paragraph(paragraph: FilesParagraph) -> str | list[str]:
    return _collapse_list_if_one_item(
        [_convert_asterisk(path) for path in list(paragraph.files)]
    )


def _comment_from_paragraph(paragraph: FilesParagraph) -> str | None:
    return cast(Optional[str], paragraph.comment)


def _annotations_from_paragraphs(
    paragraphs: Iterable[FilesParagraph],
) -> list[dict[str, str | list[str]]]:
    annotations = []
    for paragraph in paragraphs:
        copyrights = _copyrights_from_paragraph(paragraph)
        paths = _paths_from_paragraph(paragraph)
        paragraph_result = {
            "path": paths,
            "precedence": "aggregate",
            "SPDX-FileCopyrightText": copyrights,
            "SPDX-License-Identifier": paragraph.license.to_str(),
        }
        comment = _comment_from_paragraph(paragraph)
        if comment:
            paragraph_result["SPDX-FileComment"] = comment
        annotations.append(paragraph_result)
    return annotations


def toml_from_dep5(dep5: Copyright) -> str:
    """Given a Copyright object, return an equivalent REUSE.toml string."""
    header = _header_from_dep5_header(dep5.header)
    annotations = _annotations_from_paragraphs(dep5.all_files_paragraphs())
    result: dict[str, Any] = {"version": REUSE_TOML_VERSION}
    result.update(header)
    result["annotations"] = annotations
    return tomlkit.dumps(result)



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2023 DB Systel GmbH
# SPDX-FileCopyrightText: 2024 Nico Rikken <nico@nicorikken.eu>
# SPDX-FileCopyrightText: 2025 ASCIIMoth <ascii@moth.contact>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# pylint: disable=unused-argument

"""Click code for lint subcommand."""

import sys

import click

from .. import __REUSE_version__
from ..i18n import _
from ..lint import format_json, format_lines, format_plain
from ..report import ProjectReport
from .common import ClickObj, MutexOption
from .main import main

_OUTPUT_MUTEX = ["quiet", "json", "plain", "lines"]

_HELP = (
    _(
        "Lint the project directory for REUSE compliance. This version of the"
        " tool checks against version {reuse_version} of the REUSE"
        " Specification. You can find the latest version of the specification"
        " at <https://reuse.software/spec/>."
    ).format(reuse_version=__REUSE_version__)
    + "\n\n"
    + _("Specifically, the following criteria are checked:")
    + "\n\n\b\n"
    + click.wrap_text(
        _(
            "- Are there any bad (unrecognised, not compliant with SPDX)"
            " licenses in the project?"
        )
    )
    + "\n"
    + click.wrap_text(_("- Are there any deprecated licenses in the project?"))
    + "\n"
    + click.wrap_text(
        _(
            "- Are there any license files in the LICENSES/ directory"
            " without file extension?"
        )
    )
    + "\n"
    + click.wrap_text(
        _(
            "- Are any licenses referred to inside of the project, but"
            " not included in the LICENSES/ directory?"
        )
    )
    + "\n"
    + click.wrap_text(
        _(
            "- Are any licenses included in the LICENSES/ directory that"
            " are not used inside of the project?"
        )
    )
    + "\n"
    + click.wrap_text(_("- Are there any read errors?"))
    + "\n"
    + click.wrap_text(
        _("- Do all files have valid copyright and licensing information?")
    )
)


@main.command(name="lint", help=_HELP)
@click.option(
    "--quiet",
    "-q",
    cls=MutexOption,
    mutually_exclusive=_OUTPUT_MUTEX,
    is_flag=True,
    help=_("Prevent output."),
)
@click.option(
    "--json",
    "-j",
    cls=MutexOption,
    mutually_exclusive=_OUTPUT_MUTEX,
    is_flag=True,
    help=_("Format output as JSON."),
)
@click.option(
    "--plain",
    "-p",
    cls=MutexOption,
    mutually_exclusive=_OUTPUT_MUTEX,
    is_flag=True,
    help=_("Format output as plain text. (default)"),
)
@click.option(
    "--lines",
    "-l",
    cls=MutexOption,
    mutually_exclusive=_OUTPUT_MUTEX,
    is_flag=True,
    help=_("Format output as errors per line."),
)
@click.pass_obj
def lint(
    obj: ClickObj, quiet: bool, json: bool, plain: bool, lines: bool
) -> None:
    # pylint: disable=missing-function-docstring
    report = ProjectReport.generate(
        obj.project,
        do_checksum=False,
        multiprocessing=not obj.no_multiprocessing,
    )

    if quiet:
        pass
    elif json:
        click.echo(format_json(report), nl=False)
    elif lines:
        click.echo(format_lines(report), nl=False)
    else:
        click.echo(format_plain(report), nl=False)

    sys.exit(0 if report.is_compliant else 1)



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2021 Alliander N.V.
# SPDX-FileCopyrightText: 2023 Carmen Bianca BAKKER <carmenbianca@fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""reuse is a tool for compliance with the REUSE recommendations.

Although the API is documented, it is **NOT** guaranteed stable between minor or
even patch releases. The semantic versioning of this program pertains
exclusively to the reuse CLI command. If you want to use reuse as a Python
library, you should pin reuse to an exact version.

Having given the above disclaimer, the API has been relatively stable
nevertheless, and we (the maintainers) do make some efforts to not needlessly
change the public API.
"""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("reuse")
except PackageNotFoundError:
    # package is not installed
    __version__ = "6.2.0"

__author__ = "Carmen Bianca Bakker"
__email__ = "carmenbianca@fsfe.org"
__license__ = "Apache-2.0 AND CC0-1.0 AND CC-BY-SA-4.0 AND GPL-3.0-or-later"
__REUSE_version__ = "3.3"



# SPDX-FileCopyrightText: 2019 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2019 Kirill Elagin
# SPDX-FileCopyrightText: 2020 Dmitry Bogatov
# SPDX-FileCopyrightText: 2021 Alvar Penning
# SPDX-FileCopyrightText: 2021 Matija Šuklje <matija@suklje.name>
# SPDX-FileCopyrightText: 2021 Robin Vobruba <hoijui.quaero@gmail.com>
# SPDX-FileCopyrightText: 2021-2022 Alliander N.V. <https://alliander.com>
# SPDX-FileCopyrightText: 2022 Carmen Bianca Bakker <carmenbianca@fsfe.org>
# SPDX-FileCopyrightText: 2022, 2025 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2022 Nico Rikken <nico.rikken@fsfe.org>
# SPDX-FileCopyrightText: 2022 Sebastian Crane <seabass@fsfe.org>
# SPDX-FileCopyrightText: 2022 Stefan Hynek <stefan.hynek@uni-goettingen.de>
# SPDX-FileCopyrightText: 2022 Arnouet Engelen
# SPDX-FileCopyrightText: 2023 Juelich Supercomputing Centre, Forschungszentrum Juelich GmbH
# SPDX-FileCopyrightText: 2023 Kevin Meagher
# SPDX-FileCopyrightText: 2023 Mathias Dannesbo <md@magenta.dk>
# SPDX-FileCopyrightText: 2023 Maxim Cournoyer <maxim.cournoyer@gmail.com>
# SPDX-FileCopyrightText: 2023 Redradix S.L. <info@redradix.com>
# SPDX-FileCopyrightText: 2023 Shun Sakai <sorairolake@protonmail.ch>
# SPDX-FileCopyrightText: 2024 Rivos Inc.
# SPDX-FileCopyrightText: 2024 Anthony Loiseau <anthony.loiseau@allcircuits.com>
# SPDX-FileCopyrightText: 2024 Yongmin Hong <yewon@revi.email>
# SPDX-FileCopyrightText: 2025 Raphael Schlarb <info@raphael.schlarb.one>
# SPDX-FileCopyrightText: 2025 András Nagy <nagyandris0718@gmail.com>
# SPDX-FileCopyrightText: 2025 Kiko Fernandez-Reyes <kiko@erlang.org>
# SPDX-FileCopyrightText: 2025 Hatzka <hatzka@nezumi.studio>
# SPDX-FileCopyrightText: 2025 Johannes HAMPP <johannes.hampp@openenergytransition.org>
# SPDX-FileCopyrightText: 2025 Simon Barth <simon.barth@gmx.de>
# SPDX-FileCopyrightText: 2025 Maximilian Franzke <mfr@nzke.net>
# SPDX-FileCopyrightText: 2025 Sacha-Élie Ayoun <sachaayoun@gmail.com>
# SPDX-FileCopyrightText: 2025 Thomas Gilon <thomas.gilon@openenergytransition.org>
# SPDX-FileCopyrightText: 2025 Manlio Perillo <manlio.perillo@gmail.com>
# SPDX-FileCopyrightText: 2025 Matthias Schoettle <opensource@mattsch.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# pylint: disable=too-many-lines

"""Module for parsing and creating comments. Just enough to deal with comment
headers, in any case.
"""

import logging
import operator
import re
from pathlib import Path
from textwrap import dedent
from typing import NamedTuple

from .exceptions import CommentCreateError, CommentParseError
from .types import StrPath

_LOGGER = logging.getLogger(__name__)


class MultiLineSegments(NamedTuple):
    """Components that make up a multi-line comment style, e.g. ``'/*'``,
    ``'*'``, and ``'*/'``.
    """

    start: str
    middle: str
    end: str


class CommentStyle:
    """Base class for comment style."""

    SHORTHAND = ""
    SINGLE_LINE = ""
    SINGLE_LINE_REGEXP: re.Pattern | None = None
    INDENT_AFTER_SINGLE = ""
    # (start, middle, end)
    # e.g., ("/*", "*", "*/")
    MULTI_LINE = MultiLineSegments("", "", "")
    INDENT_BEFORE_MIDDLE = ""
    INDENT_AFTER_MIDDLE = ""
    INDENT_BEFORE_END = ""
    SHEBANGS: list[str] = []

    @classmethod
    def can_handle_single(cls) -> bool:
        """Whether the :class:`CommentStyle` can handle single-line comments."""
        return bool(cls.SINGLE_LINE)

    @classmethod
    def can_handle_multi(cls) -> bool:
        """Whether the :class:`CommentStyle` can handle multi-line comments."""
        return all((cls.MULTI_LINE.start, cls.MULTI_LINE.end))

    @classmethod
    def create_comment(cls, text: str, force_multi: bool = False) -> str:
        """Comment all lines in *text*. Single-line comments are preferred over
        multi-line comments, unless *force_multi* is provided.

        Raises:
            CommentCreateError: if *text* could not be commented.
        """
        if force_multi or not cls.can_handle_single():
            return cls._create_comment_multi(text)
        return cls._create_comment_single(text)

    @classmethod
    def _create_comment_single(cls, text: str) -> str:
        """Comment all lines in *text*, using single-line comments.

        Raises:
            CommentCreateError: if *text* could not be commented.
        """
        if not cls.can_handle_single():
            raise CommentCreateError(
                f"{cls} cannot create single-line comments"
            )
        result = []
        for line in text.split("\n"):
            line_result = cls.SINGLE_LINE
            if line:
                line_result += cls.INDENT_AFTER_SINGLE + line
            result.append(line_result)
        return "\n".join(result)

    @classmethod
    def _create_comment_multi(cls, text: str) -> str:
        """Comment all lines in *text*, using multi-line comments.

        Raises:
            CommentCreateError: if *text* could not be commented.
        """
        if not cls.can_handle_multi():
            raise CommentCreateError(f"{cls} cannot create multi-line comments")
        result = []
        result.append(cls.MULTI_LINE.start)
        for line in text.split("\n"):
            if cls.MULTI_LINE.end in text:
                raise CommentCreateError(
                    f"'{line}' contains a premature comment delimiter"
                )
            line_result = ""
            if cls.MULTI_LINE.middle:
                line_result += cls.INDENT_BEFORE_MIDDLE + cls.MULTI_LINE.middle
            if line:
                line_result += cls.INDENT_AFTER_MIDDLE + line
            result.append(line_result)
        result.append(cls.INDENT_BEFORE_END + cls.MULTI_LINE.end)
        return "\n".join(result)

    @classmethod
    def parse_comment(cls, text: str) -> str:
        """Uncomment all lines in *text*.

        Raises:
            CommentParseError: if *text* could not be parsed.
        """
        try:
            # Attempt to parse multi-line comments first, in case of comment
            # styles like Julia, where '#=' starts a multi-line comment, and '#'
            # starts a single-line comment. If we parsed single-line comments
            # first, '#=' would be a valid single-line comment.
            return cls._parse_comment_multi(text)
        except CommentParseError:
            return cls._parse_comment_single(text)

    @classmethod
    def _parse_comment_single(cls, text: str) -> str:
        """Uncomment all lines in *text*, assuming they are commented by
        single-line comments.

        Raises:
            CommentParseError: if *text* could not be parsed.
        """
        if not cls.can_handle_single():
            raise CommentParseError(f"{cls} cannot parse single-line comments")
        result_lines = []

        for line in text.splitlines():
            if cls.SINGLE_LINE_REGEXP:
                if match := cls.SINGLE_LINE_REGEXP.match(line):
                    line = line.removeprefix(match.group(0))
                    result_lines.append(line)
                    continue

            if not line.startswith(cls.SINGLE_LINE):
                raise CommentParseError(
                    f"'{line}' does not start with a comment marker"
                )
            line = line.removeprefix(cls.SINGLE_LINE)
            result_lines.append(line)

        result = "\n".join(result_lines)
        return dedent(result)

    @classmethod
    def _remove_middle_marker(cls, line: str) -> str:
        if cls.MULTI_LINE.middle:
            possible_line = line.lstrip()
            prefix = cls.MULTI_LINE.middle
            if possible_line.startswith(prefix):
                line = possible_line.lstrip(prefix)
                # Note to future self: line.removeprefix would be preferable
                # here.
                if line.startswith(cls.INDENT_AFTER_MIDDLE):
                    line = line.replace(cls.INDENT_AFTER_MIDDLE, "", 1)
            else:
                _LOGGER.debug(
                    "'%s' does not contain a middle comment marker", line
                )
        return line

    @classmethod
    def _parse_comment_multi(cls, text: str) -> str:
        """Uncomment all lines in *text*, assuming they are commented by
        multi-line comments.

        Raises:
            CommentParseError: if *text* could not be parsed.
        """
        if not cls.can_handle_multi():
            raise CommentParseError(f"{cls} cannot parse multi-line comments")

        result_lines = []
        try:
            first, *lines, last = text.splitlines()
            last_is_first = False
        except ValueError:
            first = text
            lines = []
            last = ""  # Set this later.
            last_is_first = True

        if not first.startswith(cls.MULTI_LINE.start):
            raise CommentParseError(
                f"'{first}' does not start with a comment marker"
            )
        first = first.lstrip(cls.MULTI_LINE.start)
        first = first.lstrip()

        for line in lines:
            line = cls._remove_middle_marker(line)
            result_lines.append(line)

        if last_is_first:
            last = first
            first = ""
        if not last.endswith(cls.MULTI_LINE.end):
            raise CommentParseError(
                f"'{last}' does not end with a comment delimiter"
            )
        last = last.rstrip(cls.MULTI_LINE.end)
        last = last.rstrip()
        last = cls._remove_middle_marker(last)

        result = "\n".join(result_lines)
        result = dedent(result)

        return "\n".join(item for item in (first, result, last) if item)

    @classmethod
    def comment_at_first_character(cls, text: str) -> str:
        """Return the comment block that starts at the first character of
        *text*. This is chiefly handy to get the header comment of a file,
        assuming that the header comment starts at the first character in the
        file.

        Raises:
            CommentParseError: if *text* does not start with a parseable
                comment block.
        """
        if not any((cls.can_handle_single(), cls.can_handle_multi())):
            raise CommentParseError(f"{cls} cannot parse comments")

        lines = text.splitlines()
        end: int | None = None

        if cls.can_handle_single():
            for i, line in enumerate(lines):
                if (
                    cls.SINGLE_LINE_REGEXP
                    and cls.SINGLE_LINE_REGEXP.match(line)
                ) or line.startswith(cls.SINGLE_LINE):
                    end = i
                else:
                    break
        if (
            end is None
            and cls.can_handle_multi()
            and text.startswith(cls.MULTI_LINE.start)
        ):
            for i, line in enumerate(lines):
                end = i
                if line.endswith(cls.MULTI_LINE.end):
                    break
            else:
                raise CommentParseError("Comment block never delimits")

        if end is not None:
            return "\n".join(lines[: end + 1])

        raise CommentParseError(
            "Could not find a parseable comment block at the first character"
        )


class AppleScriptCommentStyle(CommentStyle):
    """AppleScript comment style."""

    SHORTHAND = "applescript"

    SINGLE_LINE = "--"
    INDENT_AFTER_SINGLE = " "
    MULTI_LINE = MultiLineSegments("(*", "", "*)")


class AspxCommentStyle(CommentStyle):
    """ASPX comment style."""

    SHORTHAND = "aspx"

    MULTI_LINE = MultiLineSegments("<%--", "", "--%>")


class BatchFileCommentStyle(CommentStyle):
    """Windows batch file comment style."""

    SHORTHAND = "bat"

    SINGLE_LINE = "REM"
    INDENT_AFTER_SINGLE = " "


class BibTexCommentStyle(CommentStyle):
    """BibTex comment style."""

    SHORTHAND = "bibtex"

    MULTI_LINE = MultiLineSegments("@Comment{", "", "}")
    SHEBANGS = ["% !BIB", "%!BIB"]


class BladeCommentStyle(CommentStyle):
    """Laravel Blade Template comment style."""

    _shorthand = "blade"

    MULTI_LINE = MultiLineSegments("{{--", "", "--}}")


class CCommentStyle(CommentStyle):
    """C comment style."""

    SHORTHAND = "c"

    MULTI_LINE = MultiLineSegments("/*", "*", "*/")
    INDENT_BEFORE_MIDDLE = " "
    INDENT_AFTER_MIDDLE = " "
    INDENT_BEFORE_END = " "


class CppCommentStyle(CommentStyle):
    """C++ comment style."""

    SHORTHAND = "cpp"

    SINGLE_LINE = "//"
    INDENT_AFTER_SINGLE = " "
    MULTI_LINE = MultiLineSegments("/*", "*", "*/")
    INDENT_BEFORE_MIDDLE = " "
    INDENT_AFTER_MIDDLE = " "
    INDENT_BEFORE_END = " "
    SHEBANGS = [
        "#!",  #  V-Lang
        "<?php",  # PHP
    ]


class CppSingleCommentStyle(CommentStyle):
    """C++ single-only comment style."""

    SHORTHAND = "cppsingle"

    SINGLE_LINE = "//"
    INDENT_AFTER_SINGLE = " "
    SHEBANGS = ["#!"]  # Gleam


class EmptyCommentStyle(CommentStyle):
    """Hacky comment style for files that have no comments."""

    @classmethod
    def create_comment(cls, text: str, force_multi: bool = False) -> str:
        return text

    @classmethod
    def parse_comment(cls, text: str) -> str:
        return text

    @classmethod
    def comment_at_first_character(cls, text: str) -> str:
        return text


class FortranCommentStyle(CommentStyle):
    """Fortran (fixed form) comment style."""

    SHORTHAND = "f"

    SINGLE_LINE = "c"
    INDENT_AFTER_SINGLE = " "


class ModernFortranCommentStyle(CommentStyle):
    """Fortran (free form) comment style."""

    SHORTHAND = "f90"

    SINGLE_LINE = "!"
    INDENT_AFTER_SINGLE = " "


class FtlCommentStyle(CommentStyle):
    """FreeMarker Template Language comment style."""

    SHORTHAND = "ftl"

    MULTI_LINE = MultiLineSegments("<#--", "", "-->")


class HandlebarsCommentStyle(CommentStyle):
    """Handlebars comment style."""

    SHORTHAND = "handlebars"

    MULTI_LINE = MultiLineSegments("{{!--", "", "--}}")


class HaskellCommentStyle(CommentStyle):
    """Haskell comment style."""

    SHORTHAND = "haskell"

    SINGLE_LINE = "--"
    INDENT_AFTER_SINGLE = " "
    SHEBANGS = ["cabal-version:"]


class HtmlCommentStyle(CommentStyle):
    """HTML comment style."""

    SHORTHAND = "html"

    MULTI_LINE = MultiLineSegments("<!--", "", "-->")
    SHEBANGS = ["<?xml"]


class JinjaCommentStyle(CommentStyle):
    """Jinja2 comment style."""

    SHORTHAND = "jinja"

    MULTI_LINE = MultiLineSegments("{#", "", "#}")


class JuliaCommentStyle(CommentStyle):
    """Julia comment style."""

    SHORTHAND = "julia"

    SINGLE_LINE = "#"
    INDENT_AFTER_SINGLE = " "
    MULTI_LINE = MultiLineSegments("#=", "", "=#")
    SHEBANGS = ["#!"]


class LeanCommentStyle(CommentStyle):
    """Lean comment style."""

    SHORTHAND = "lean"

    SINGLE_LINE = "--"
    INDENT_AFTER_SINGLE = " "
    MULTI_LINE = MultiLineSegments("/-", "-", "-/")
    INDENT_BEFORE_MIDDLE = " "
    INDENT_AFTER_MIDDLE = " "
    INDENT_BEFORE_END = " "


class LispCommentStyle(CommentStyle):
    """Lisp comment style."""

    SHORTHAND = "lisp"

    SINGLE_LINE = ";;;"
    SINGLE_LINE_REGEXP = re.compile(r"^;+\s*")
    INDENT_AFTER_SINGLE = " "


class M4CommentStyle(CommentStyle):
    """M4 (autoconf) comment style."""

    SHORTHAND = "m4"

    SINGLE_LINE = "dnl"
    INDENT_AFTER_SINGLE = " "


class MlCommentStyle(CommentStyle):
    """ML comment style."""

    SHORTHAND = "ml"

    MULTI_LINE = MultiLineSegments("(*", "*", "*)")
    INDENT_BEFORE_MIDDLE = " "
    INDENT_AFTER_MIDDLE = " "
    INDENT_BEFORE_END = " "


class PlantUmlCommentStyle(CommentStyle):
    """PlantUML comment style."""

    SHORTHAND = "plantuml"

    SINGLE_LINE = "'"
    INDENT_AFTER_SINGLE = " "
    MULTI_LINE = MultiLineSegments("/'", "'", "'/")
    INDENT_BEFORE_MIDDLE = " "
    INDENT_AFTER_MIDDLE = " "
    INDENT_BEFORE_END = " "


class PythonCommentStyle(CommentStyle):
    """Python comment style."""

    SHORTHAND = "python"

    SINGLE_LINE = "#"
    INDENT_AFTER_SINGLE = " "
    SHEBANGS = ["#!"]


class ReStructedTextCommentStyle(CommentStyle):
    """reStructuredText comment style."""

    SHORTHAND = "rst"

    SINGLE_LINE = ".."
    INDENT_AFTER_SINGLE = " "


class SemicolonCommentStyle(CommentStyle):
    """Semicolon comment style."""

    SHORTHAND = "semicolon"

    SINGLE_LINE = ";"
    INDENT_AFTER_SINGLE = " "


class TexCommentStyle(CommentStyle):
    """TeX comment style."""

    SHORTHAND = "tex"

    SINGLE_LINE = "%"
    INDENT_AFTER_SINGLE = " "
    SHEBANGS = ["% !TEX", "%!TEX", "#!"]


class UncommentableCommentStyle(EmptyCommentStyle):
    """A pseudo comment style to indicate that this file is uncommentable. This
    results in an external .license file for binaries and --force-dot-license.
    """


class UnixManCommentStyle(CommentStyle):
    """UNIX manual page comment style."""

    SHORTHAND = "man"

    # In case the below is difficult to read, the comment character is: .\"
    SINGLE_LINE = r".\""
    INDENT_AFTER_SINGLE = " "


class VelocityCommentStyle(CommentStyle):
    """Apache Velocity Template Language comment style."""

    SHORTHAND = "vst"

    # TODO: SINGLE_LINE requires refactor to support trailing `**`.
    MULTI_LINE = MultiLineSegments("#*", "  ", "*#")


class VimCommentStyle(CommentStyle):
    """Vim(Script|Config) style."""

    SHORTHAND = "vim"

    SINGLE_LINE = '"'
    INDENT_AFTER_SINGLE = " "


class XQueryCommentStyle(CommentStyle):
    """XQuery comment style."""

    SHORTHAND = "xquery"

    MULTI_LINE = MultiLineSegments("(:", ":", ":)")
    INDENT_BEFORE_MIDDLE = " "
    INDENT_AFTER_MIDDLE = " "
    INDENT_BEFORE_END = " "


#: A map of (common) file extensions against comment types.
EXTENSION_COMMENT_STYLE_MAP = {
    ".adb": HaskellCommentStyle,
    ".adoc": CppCommentStyle,
    ".ads": HaskellCommentStyle,
    ".aes": UncommentableCommentStyle,
    ".ahk": SemicolonCommentStyle,
    ".ahkl": SemicolonCommentStyle,
    ".aidl": CppCommentStyle,
    ".applescript": AppleScriptCommentStyle,
    ".arb": UncommentableCommentStyle,
    ".asax": AspxCommentStyle,
    ".asc": CppCommentStyle,
    ".asciidoc": CppCommentStyle,
    ".ashx": AspxCommentStyle,
    ".asm": LispCommentStyle,  # ASM assembler
    ".asmx": AspxCommentStyle,
    ".aspx": AspxCommentStyle,
    ".aux": TexCommentStyle,
    ".awk": PythonCommentStyle,
    ".axd": AspxCommentStyle,
    ".bash": PythonCommentStyle,
    ".bat": BatchFileCommentStyle,
    ".bats": PythonCommentStyle,
    ".bb": PythonCommentStyle,
    ".bbappend": PythonCommentStyle,
    ".bbclass": PythonCommentStyle,
    ".bib": BibTexCommentStyle,
    ".blade.php": BladeCommentStyle,
    ".bzl": PythonCommentStyle,
    ".c": CCommentStyle,
    ".cabal": HaskellCommentStyle,
    ".cc": CppCommentStyle,
    ".cjs": CppCommentStyle,
    ".cl": LispCommentStyle,
    ".clj": LispCommentStyle,
    ".cljc": LispCommentStyle,
    ".cljs": LispCommentStyle,
    ".cls": TexCommentStyle,
    ".cmake": PythonCommentStyle,  # TODO: Bracket comments not supported.
    ".code-workspace": CppCommentStyle,
    ".coffee": PythonCommentStyle,
    ".cpp": CppCommentStyle,
    ".cs": CppCommentStyle,
    ".csl": HtmlCommentStyle,  # Bibliography (XML based)
    ".cson": PythonCommentStyle,
    ".css": CCommentStyle,
    ".csproj": HtmlCommentStyle,
    ".csv": UncommentableCommentStyle,
    ".cu": CppCommentStyle,
    ".cuh": CppCommentStyle,
    ".cxx": CppCommentStyle,
    ".d": CppCommentStyle,
    ".dart": CppCommentStyle,
    ".di": CppCommentStyle,
    ".doc": UncommentableCommentStyle,
    ".docx": UncommentableCommentStyle,
    ".dotx": UncommentableCommentStyle,
    ".dtd": HtmlCommentStyle,
    ".dts": CppCommentStyle,
    ".dtsi": CppCommentStyle,
    ".el": LispCommentStyle,
    ".erl": TexCommentStyle,
    ".escript": TexCommentStyle,
    ".es": TexCommentStyle,
    ".ex": PythonCommentStyle,
    ".exs": PythonCommentStyle,
    ".f": FortranCommentStyle,
    ".fsproj": HtmlCommentStyle,
    ".f03": ModernFortranCommentStyle,
    ".f08": ModernFortranCommentStyle,
    ".f90": ModernFortranCommentStyle,
    ".f95": ModernFortranCommentStyle,
    ".fish": PythonCommentStyle,
    ".fnl": LispCommentStyle,
    ".fodp": UncommentableCommentStyle,
    ".fods": UncommentableCommentStyle,
    ".fodt": UncommentableCommentStyle,
    ".for": FortranCommentStyle,
    ".ftn": FortranCommentStyle,
    ".fpp": FortranCommentStyle,
    ".fs": CppCommentStyle,
    ".fsx": CppCommentStyle,
    ".ftl": FtlCommentStyle,
    ".gemspec": PythonCommentStyle,
    ".gleam": CppSingleCommentStyle,
    ".go": CppCommentStyle,
    ".gperf": CppCommentStyle,
    ".gradle": CppCommentStyle,
    ".graphql": PythonCommentStyle,
    ".graphqls": PythonCommentStyle,
    ".gqls": PythonCommentStyle,
    ".groovy": CppCommentStyle,
    ".h": CCommentStyle,
    ".ha": CppSingleCommentStyle,
    ".hbs": HandlebarsCommentStyle,
    ".hcl": PythonCommentStyle,
    ".hh": CppCommentStyle,
    ".hjson": CppCommentStyle,
    ".hpp": CppCommentStyle,
    ".hrl": TexCommentStyle,
    ".hs": HaskellCommentStyle,
    ".html": HtmlCommentStyle,
    ".hx": CppCommentStyle,
    ".hxsl": CppCommentStyle,
    ".ilean": UncommentableCommentStyle,
    ".ini": SemicolonCommentStyle,
    ".ino": CppCommentStyle,
    ".ipynb": UncommentableCommentStyle,
    ".iuml": PlantUmlCommentStyle,
    ".j2": JinjaCommentStyle,
    ".java": CppCommentStyle,
    ".jinja": JinjaCommentStyle,
    ".jinja2": JinjaCommentStyle,
    ".jl": JuliaCommentStyle,
    ".jpg": UncommentableCommentStyle,
    ".jpeg": UncommentableCommentStyle,
    ".js": CppCommentStyle,
    ".json": UncommentableCommentStyle,
    ".json5": CppCommentStyle,
    ".jsonc": CppCommentStyle,
    ".jsp": AspxCommentStyle,
    ".jsx": CppCommentStyle,
    ".jy": PythonCommentStyle,
    ".ksh": PythonCommentStyle,
    ".kt": CppCommentStyle,
    ".kts": CppCommentStyle,
    ".l": LispCommentStyle,
    ".latex": TexCommentStyle,
    ".ld": CCommentStyle,
    ".lean": LeanCommentStyle,
    ".less": CCommentStyle,
    ".license": EmptyCommentStyle,
    ".lisp": LispCommentStyle,
    ".lsp": LispCommentStyle,
    ".lua": HaskellCommentStyle,
    ".m": TexCommentStyle,
    ".m4": M4CommentStyle,
    ".man": UnixManCommentStyle,
    ".markdown": HtmlCommentStyle,
    ".md": HtmlCommentStyle,
    ".mjs": CppCommentStyle,
    ".mk": PythonCommentStyle,
    ".ml": MlCommentStyle,
    ".mli": MlCommentStyle,
    ".nim": PythonCommentStyle,
    ".nim.cfg": PythonCommentStyle,  # Nim-lang build config parameters/settings
    ".nimble": PythonCommentStyle,  # Nim-lang build config
    ".nimrod": PythonCommentStyle,
    ".nix": PythonCommentStyle,
    ".odb": UncommentableCommentStyle,
    ".odf": UncommentableCommentStyle,
    ".odg": UncommentableCommentStyle,
    ".odm": UncommentableCommentStyle,
    ".odp": UncommentableCommentStyle,
    ".ods": UncommentableCommentStyle,
    ".odt": UncommentableCommentStyle,
    ".olean": UncommentableCommentStyle,
    ".org": PythonCommentStyle,
    ".otp": UncommentableCommentStyle,
    ".ots": UncommentableCommentStyle,
    ".ott": UncommentableCommentStyle,
    ".pdf": UncommentableCommentStyle,
    ".pem": UncommentableCommentStyle,
    ".php": CppCommentStyle,
    ".php3": CppCommentStyle,
    ".php4": CppCommentStyle,
    ".php5": CppCommentStyle,
    ".pl": PythonCommentStyle,
    ".plantuml": PlantUmlCommentStyle,
    ".png": UncommentableCommentStyle,
    ".po": PythonCommentStyle,
    ".pod": PythonCommentStyle,
    ".pot": PythonCommentStyle,
    ".ppt": UncommentableCommentStyle,
    ".pptx": UncommentableCommentStyle,
    ".pri": PythonCommentStyle,
    ".pro": PythonCommentStyle,
    ".props": HtmlCommentStyle,  # MSBuild files
    ".properties": PythonCommentStyle,
    ".proto": CppCommentStyle,
    ".ps1": PythonCommentStyle,  # TODO: Multiline comments
    ".psm1": PythonCommentStyle,  # TODO: Multiline comments
    ".pu": PlantUmlCommentStyle,
    ".puml": PlantUmlCommentStyle,
    ".pxd": PythonCommentStyle,
    ".py": PythonCommentStyle,
    ".pyi": PythonCommentStyle,
    ".pyw": PythonCommentStyle,
    ".pyx": PythonCommentStyle,
    ".qbs": CppCommentStyle,
    ".qml": CppCommentStyle,
    ".qrc": HtmlCommentStyle,
    ".qss": CCommentStyle,
    ".R": PythonCommentStyle,
    ".rake": PythonCommentStyle,
    ".rb": PythonCommentStyle,
    ".rbw": PythonCommentStyle,
    ".rbx": PythonCommentStyle,
    ".rkt": LispCommentStyle,
    ".Rmd": HtmlCommentStyle,
    ".rs": CppCommentStyle,
    ".rss": HtmlCommentStyle,
    ".rst": ReStructedTextCommentStyle,
    ".s": CCommentStyle,
    ".sass": CCommentStyle,
    ".sbt": CppCommentStyle,
    ".sc": CppCommentStyle,  # SuperCollider source file
    ".scad": CppCommentStyle,
    ".scala": CppCommentStyle,
    ".scm": LispCommentStyle,
    ".scpt": AppleScriptCommentStyle,
    ".scptd": AppleScriptCommentStyle,
    ".scss": CCommentStyle,
    # SuperCollider synth definition (binary)
    ".scsyndef": UncommentableCommentStyle,
    ".sh": PythonCommentStyle,
    ".sld": LispCommentStyle,  # Scheme Library Definition (R7RS)
    # Visual Studio solution file, officially uncommentable:
    ".sln": UncommentableCommentStyle,
    ".sls": LispCommentStyle,  # Scheme Library Source (R6RS)
    ".smk": PythonCommentStyle,
    ".sml": MlCommentStyle,
    ".soy": CppCommentStyle,
    ".sps": LispCommentStyle,  # Scheme Program Source (R6RS)
    ".sql": HaskellCommentStyle,
    ".sty": TexCommentStyle,
    ".svg": UncommentableCommentStyle,
    ".svelte": HtmlCommentStyle,
    ".swift": CppCommentStyle,
    ".t": PythonCommentStyle,
    ".tcl": PythonCommentStyle,
    ".tex": TexCommentStyle,
    ".textile": HtmlCommentStyle,
    ".tf": PythonCommentStyle,
    ".tfvars": PythonCommentStyle,
    ".thy": MlCommentStyle,
    ".toc": TexCommentStyle,
    ".toml": PythonCommentStyle,
    ".ts": CppCommentStyle,
    ".tsx": CppCommentStyle,
    ".ttl": PythonCommentStyle,  # Turtle/RDF
    ".typ": CppCommentStyle,  # typst files
    ".ui": HtmlCommentStyle,
    ".v": CppCommentStyle,  # V-Lang source code
    ".vala": CppCommentStyle,
    ".vbproj": HtmlCommentStyle,
    ".vhdl": HaskellCommentStyle,
    ".vim": VimCommentStyle,
    ".vm": VelocityCommentStyle,
    ".vsh": CppCommentStyle,  # V-Lang script
    ".vtl": VelocityCommentStyle,
    ".vue": HtmlCommentStyle,
    ".webp": UncommentableCommentStyle,
    ".xls": UncommentableCommentStyle,
    ".xlsx": UncommentableCommentStyle,
    ".xml": HtmlCommentStyle,
    ".xq": XQueryCommentStyle,
    ".xql": XQueryCommentStyle,
    ".xqm": XQueryCommentStyle,
    ".xqy": XQueryCommentStyle,
    ".xquery": XQueryCommentStyle,
    ".xrl": TexCommentStyle,
    ".xsd": HtmlCommentStyle,
    ".xsh": PythonCommentStyle,
    ".xsl": HtmlCommentStyle,
    ".yaml": PythonCommentStyle,
    ".yml": PythonCommentStyle,
    ".yrl": TexCommentStyle,
    ".zig": CppSingleCommentStyle,
    ".zsh": PythonCommentStyle,
}

EXTENSION_COMMENT_STYLE_MAP_LOWERCASE = {
    key.lower(): value for key, value in EXTENSION_COMMENT_STYLE_MAP.items()
}

FILENAME_COMMENT_STYLE_MAP = {
    ".arcconfig": UncommentableCommentStyle,  # is a JSON file
    ".arclint": UncommentableCommentStyle,  # is a JSON file
    ".arcunit": UncommentableCommentStyle,  # is a JSON file
    ".bashrc": PythonCommentStyle,
    ".bazelignore": PythonCommentStyle,
    ".bazelrc": PythonCommentStyle,
    ".browserslist": PythonCommentStyle,
    ".clang-format": PythonCommentStyle,
    ".clang-tidy": PythonCommentStyle,
    ".coveragerc": PythonCommentStyle,
    ".dockerignore": PythonCommentStyle,
    ".earthlyignore": PythonCommentStyle,
    ".editorconfig": PythonCommentStyle,
    ".envrc": PythonCommentStyle,
    ".empty": EmptyCommentStyle,
    ".eslintignore": PythonCommentStyle,
    ".eslintrc": UncommentableCommentStyle,
    ".gitattributes": PythonCommentStyle,
    ".gitignore": PythonCommentStyle,
    ".gitmodules": PythonCommentStyle,
    ".htaccess": PythonCommentStyle,
    ".mailmap": PythonCommentStyle,
    ".metadata": UncommentableCommentStyle,
    ".mdlrc": PythonCommentStyle,  # Markdown-linter config
    ".npmignore": PythonCommentStyle,
    ".npmrc": SemicolonCommentStyle,
    ".nvmrc": PythonCommentStyle,
    ".prettierrc": UncommentableCommentStyle,  # could either be JSON or YAML
    ".prettierignore": PythonCommentStyle,
    ".pylintrc": PythonCommentStyle,
    ".Renviron": PythonCommentStyle,
    ".Rprofile": PythonCommentStyle,
    ".shellcheckrc": PythonCommentStyle,
    ".taprc": PythonCommentStyle,
    ".vimrc": VimCommentStyle,
    ".zshrc": PythonCommentStyle,
    ".yamllint": PythonCommentStyle,  # is a YAML file
    ".yarnrc": PythonCommentStyle,
    "ansible.cfg": PythonCommentStyle,
    "archive.sctxar": UncommentableCommentStyle,  # SuperCollider global archive
    "cabal.project": HaskellCommentStyle,
    "Cargo.lock": UncommentableCommentStyle,
    "CMakeLists.txt": PythonCommentStyle,
    "CODEOWNERS": PythonCommentStyle,
    "configure.ac": M4CommentStyle,
    "Containerfile": PythonCommentStyle,
    "Dockerfile": PythonCommentStyle,
    "Doxyfile": PythonCommentStyle,
    "dune": SemicolonCommentStyle,  # OCaml build system
    "dune-project": SemicolonCommentStyle,  # OCaml build system
    "dune-workspace": SemicolonCommentStyle,  # OCaml build system
    "Earthfile": PythonCommentStyle,
    "flake.lock": UncommentableCommentStyle,  # is a JSON file
    "Gemfile": PythonCommentStyle,
    "go.mod": CppCommentStyle,
    "go.sum": UncommentableCommentStyle,
    "gradlew": PythonCommentStyle,
    "Jenkinsfile": CppCommentStyle,
    "Makefile.am": PythonCommentStyle,
    "Makefile": PythonCommentStyle,
    "MANIFEST.in": PythonCommentStyle,
    "manifest": PythonCommentStyle,  # used by cdist
    "matplotlibrc": PythonCommentStyle,
    "meson.build": PythonCommentStyle,
    "meson_options.txt": PythonCommentStyle,
    "poetry.lock": UncommentableCommentStyle,
    "pubspec.lock": UncommentableCommentStyle,
    "pylintrc": PythonCommentStyle,
    "py.typed": PythonCommentStyle,
    "Rakefile": PythonCommentStyle,
    "requirements.txt": PythonCommentStyle,
    "ROOT": MlCommentStyle,
    "setup.cfg": PythonCommentStyle,
    "Snakefile": PythonCommentStyle,
    "uv.lock": UncommentableCommentStyle,
    "yarn.lock": UncommentableCommentStyle,
}

FILENAME_COMMENT_STYLE_MAP_LOWERCASE = {
    key.lower(): value for key, value in FILENAME_COMMENT_STYLE_MAP.items()
}


def _all_style_classes() -> list[type[CommentStyle]]:
    """Return a list of all defined style classes, excluding the base class."""
    result = []
    for key, value in globals().items():
        if key.endswith("CommentStyle") and key != "CommentStyle":
            result.append(value)
    return sorted(result, key=operator.attrgetter("__name__"))


_result = _all_style_classes()
_result.remove(EmptyCommentStyle)
_result.remove(UncommentableCommentStyle)

#: A map of human-friendly names against style classes.
NAME_STYLE_MAP = {style.SHORTHAND: style for style in _result}


def get_comment_style(path: StrPath) -> type[CommentStyle] | None:
    """Return value of CommentStyle detected for *path* or None."""
    path = Path(path)
    style = FILENAME_COMMENT_STYLE_MAP_LOWERCASE.get(path.name.lower())
    if style is None:
        style = EXTENSION_COMMENT_STYLE_MAP_LOWERCASE.get(
            "".join(path.suffixes).lower()
        )
        if style is None:
            style = EXTENSION_COMMENT_STYLE_MAP_LOWERCASE.get(
                path.suffix.lower()
            )
    return style


def is_uncommentable(path: Path) -> bool:
    """*path*'s extension has the UncommentableCommentStyle."""
    return get_comment_style(path) == UncommentableCommentStyle


def has_style(path: Path) -> bool:
    """*path*'s extension has a CommentStyle."""
    return get_comment_style(path) is not None



# SPDX-FileCopyrightText: 2019 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2023 Nico Rikken <nico.rikken@fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Click code for download subcommand."""

import logging
import sys
from collections.abc import Collection
from difflib import SequenceMatcher
from pathlib import Path
from typing import IO
from urllib.error import URLError

import click

from .._licenses import ALL_MAP, ALL_NON_DEPRECATED_MAP
from .._util import _strip_plus_from_identifier
from ..download import _path_to_license_file, put_license_in_file
from ..i18n import _
from ..report import ProjectReport
from ..types import StrPath
from .common import ClickObj, MutexOption
from .main import main

_LOGGER = logging.getLogger(__name__)


def _similar_spdx_identifiers(identifier: str) -> list[str]:
    """Given an incorrect SPDX identifier, return a list of similar ones."""
    suggestions: list[str] = []
    for valid_identifier in ALL_NON_DEPRECATED_MAP:
        distance = SequenceMatcher(
            a=identifier.lower(), b=valid_identifier[: len(identifier)].lower()
        ).ratio()
        if distance > 0.75:
            suggestions.append(valid_identifier)
    suggestions = sorted(suggestions)

    return suggestions


def _print_incorrect_spdx_identifier(
    identifier: str, out: IO[str] = sys.stdout
) -> None:
    """Print out that *identifier* is not valid, and follow up with some
    suggestions.
    """
    out.write(
        _("'{}' is not a valid SPDX License Identifier.").format(identifier)
    )
    out.write("\n")

    suggestions = _similar_spdx_identifiers(identifier)
    if suggestions:
        out.write("\n")
        out.write(_("Did you mean:"))
        out.write("\n")
        for suggestion in suggestions:
            out.write(f"* {suggestion}\n")
        out.write("\n")
    out.write(
        _(
            "See <https://spdx.org/licenses/> for a list of valid "
            "SPDX License Identifiers."
        )
    )
    out.write("\n")


def _already_exists(path: StrPath) -> None:
    click.echo(
        _("Error: {spdx_identifier} already exists.").format(
            spdx_identifier=path
        )
    )


def _not_found(path: StrPath) -> None:
    click.echo(_("Error: {path} does not exist.").format(path=path))


def _could_not_download(identifier: str) -> None:
    click.echo(_("Error: Failed to download license."))
    click.echo("")
    if identifier not in ALL_MAP:
        _print_incorrect_spdx_identifier(identifier, out=sys.stdout)
    else:
        click.echo(_("Is your internet connection working?"))


def _successfully_downloaded(destination: StrPath) -> None:
    click.echo(
        _("Successfully downloaded {spdx_identifier}.").format(
            spdx_identifier=destination
        )
    )


_ALL_MUTEX = ["all_", "output"]


_HELP = (
    _("Download a license and place it in the LICENSES/ directory.")
    + "\n\n"
    + _(
        "LICENSE must be a valid SPDX License Identifier. You may specify"
        " LICENSE multiple times to download multiple licenses."
    )
)


@main.command(name="download", help=_HELP)
@click.option(
    "--all",
    "all_",
    cls=MutexOption,
    mutually_exclusive=_ALL_MUTEX,
    is_flag=True,
    help=_("Download all missing licenses detected in the project."),
)
@click.option(
    "--output",
    "-o",
    cls=MutexOption,
    mutually_exclusive=_ALL_MUTEX,
    type=click.Path(dir_okay=False, writable=True, path_type=Path),
    help=_("Path to download to."),
)
@click.option(
    "--source",
    type=click.Path(exists=True, readable=True, path_type=Path),
    help=_(
        "Source from which to copy custom LicenseRef- licenses, either"
        " a directory that contains the file or the file itself."
    ),
)
@click.argument(
    "licenses",
    # TRANSLATORS: You may translate this. Please preserve capital letters.
    metavar=_("LICENSE"),
    type=str,
    nargs=-1,
)
@click.pass_obj
def download(
    obj: ClickObj,
    licenses: Collection[str],
    all_: bool,
    output: Path | None,
    source: Path | None,
) -> None:
    # pylint: disable=missing-function-docstring
    if all_ and licenses:
        raise click.UsageError(
            _(
                "The 'LICENSE' argument and '--all' option are mutually"
                " exclusive."
            )
        )

    if all_:
        # TODO: This is fairly inefficient, but gets the job done.
        report = ProjectReport.generate(obj.project, do_checksum=False)
        licenses = report.missing_licenses.keys()

    if len(licenses) > 1 and output:
        raise click.UsageError(
            _("Cannot use '--output' with more than one license.")
        )

    licenses = {_strip_plus_from_identifier(lic) for lic in licenses}
    return_code = 0
    for lic in licenses:
        destination: Path = output  # type: ignore
        if destination is None:
            destination = _path_to_license_file(lic, obj.project)
        try:
            put_license_in_file(lic, destination=destination, source=source)
        except URLError:
            _could_not_download(lic)
            return_code = 1
        except FileExistsError as err:
            _already_exists(err.filename)
            return_code = 1
        except FileNotFoundError as err:
            _not_found(err.filename)
            return_code = 1
        else:
            _successfully_downloaded(destination)
    sys.exit(return_code)



# SPDX-FileCopyrightText: 2019 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2023 Nico Rikken <nico.rikken@fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Functions for downloading license files from spdx/license-list-data."""

import errno
import logging
import os
import shutil
import urllib.request
from pathlib import Path
from urllib.error import URLError
from urllib.parse import urljoin

from ._licenses import ALL_NON_DEPRECATED_MAP
from ._util import find_licenses_directory
from .extract import _LICENSEREF_PATTERN
from .project import Project
from .types import StrPath
from .vcs import VCSStrategyNone

_LOGGER = logging.getLogger(__name__)

# All raw text files are available as files underneath this path.
_SPDX_REPOSITORY_BASE_URL = (
    "https://raw.githubusercontent.com/spdx/license-list-data/master/text/"
)


def download_license(spdx_identifier: str) -> str:
    """Download the license text from the SPDX repository.

    Args:
        spdx_identifier: SPDX identifier of the license.

    Raises:
        URLError: if the license could not be downloaded.

    Returns:
        The license text.
    """
    if spdx_identifier not in ALL_NON_DEPRECATED_MAP:
        spdx_identifier = f"deprecated_{spdx_identifier}"
    # This is fairly naive, but I can't see anything wrong with it.
    url = urljoin(_SPDX_REPOSITORY_BASE_URL, "".join((spdx_identifier, ".txt")))
    _LOGGER.debug("downloading license from '%s'", url)
    # TODO: Cache result?
    with urllib.request.urlopen(url) as response:
        if response.getcode() == 200:
            return response.read().decode("utf-8")
    raise URLError("Status code was not 200")


def _path_to_license_file(spdx_identifier: str, project: Project) -> Path:
    root: Path | None = project.root
    # Hack
    if (
        root
        and root.name == "LICENSES"
        and isinstance(project.vcs_strategy, VCSStrategyNone)
    ):
        root = None

    licenses_path = find_licenses_directory(root=root)
    return licenses_path / "".join((spdx_identifier, ".txt"))


def put_license_in_file(
    spdx_identifier: str,
    destination: StrPath,
    source: StrPath | None = None,
) -> None:
    """Download a license and put it in the destination file.

    This function exists solely for convenience.

    Args:
        spdx_identifier: SPDX License Identifier of the license.
        destination: Where to put the license.
        source: Path to file or directory containing the text for LicenseRef
            licenses.

    Raises:
        URLError: if the license could not be downloaded.
        FileExistsError: if the license file already exists.
        FileNotFoundError: if the source could not be found in the directory.
    """
    header = ""
    destination = Path(destination)
    destination.parent.mkdir(exist_ok=True)

    if destination.exists():
        raise FileExistsError(
            errno.EEXIST, os.strerror(errno.EEXIST), str(destination)
        )

    # LicenseRef- license; don't download anything.
    if _LICENSEREF_PATTERN.match(spdx_identifier):
        if source:
            source = Path(source)
            if source.is_dir():
                source = source / f"{spdx_identifier}.txt"
            if not source.exists():
                raise FileNotFoundError(
                    errno.ENOENT, os.strerror(errno.ENOENT), str(source)
                )
            shutil.copyfile(source, destination)
        else:
            destination.touch()
    else:
        text = download_license(spdx_identifier)
        with destination.open("w", encoding="utf-8") as fp:
            fp.write(header)
            fp.write(text)



# SPDX-FileCopyrightText: 2024 Carmen Bianca BAKKER <carmenbianca@fsfe.org>
# SPDX-FileCopyrightText: 2024 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Click code for convert-dep5 subcommand."""

from typing import cast

import click

from ..convert_dep5 import toml_from_dep5
from ..global_licensing import ReuseDep5
from ..i18n import _
from .common import ClickObj
from .main import main

_HELP = _(
    "Convert .reuse/dep5 into a REUSE.toml file. The generated file is placed"
    " in the project root and is semantically identical. The .reuse/dep5 file"
    " is subsequently deleted."
)


@main.command(name="convert-dep5", help=_HELP)
@click.pass_obj
def convert_dep5(obj: ClickObj) -> None:
    # pylint: disable=missing-function-docstring
    project = obj.project
    if not (project.root / ".reuse/dep5").exists():
        raise click.UsageError(_("No '.reuse/dep5' file."))

    text = toml_from_dep5(
        cast(ReuseDep5, project.global_licensing).dep5_copyright
    )
    (project.root / "REUSE.toml").write_text(text)
    (project.root / ".reuse/dep5").unlink()



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2022 Pietro Albini <pietro.albini@ferrous-systems.com>
# SPDX-FileCopyrightText: 2023 DB Systel GmbH
# SPDX-FileCopyrightText: 2023 Carmen Bianca BAKKER <carmenbianca@fsfe.org>
# SPDX-FileCopyrightText: 2024 Kerry McAdams <github@klmcadams>
# SPDX-FileCopyrightText: 2024 Sebastien Morais <github@SMoraisAnsys>
# SPDX-FileCopyrightText: 2025 Simon Barth <simon.barth@gmx.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Module that contains reports about files and projects for linting."""

import bdb
import contextlib
import datetime
import logging
import random
from collections import defaultdict
from collections.abc import Collection, Generator
from concurrent.futures import ProcessPoolExecutor
from functools import cached_property
from hashlib import md5
from io import StringIO
from os import cpu_count
from pathlib import Path, PurePath
from typing import Any, Final, NamedTuple, Optional, Protocol, cast
from uuid import uuid4

from . import __REUSE_version__, __version__
from ._util import (
    _add_plus_to_identifier,
    _checksum,
    _strip_plus_from_identifier,
)
from .copyright import SpdxExpression
from .extract import _LICENSEREF_PATTERN
from .global_licensing import ReuseDep5
from .i18n import _
from .project import Project, ReuseInfo
from .types import StrPath

_LOGGER = logging.getLogger(__name__)

LINT_VERSION = "1.0"

_CPU_COUNT: Final[int] = cpu_count() or 1
#: This variable exists to be able to override parallelisation. If set to
#: :const:`False`, generating :meth:`FileReport.generate` will not use
#: parallelisation.
ENABLE_PARALLEL = True

# REUSE-IgnoreStart


class _MultiprocessingContainer:
    """Container that remembers some data in order to generate a FileReport."""

    def __init__(
        self, project: Project, do_checksum: bool, add_license_concluded: bool
    ):
        if isinstance(project.global_licensing, ReuseDep5):
            # Remember that a dep5_copyright was (or was not) set prior.
            self.has_dep5 = bool(project.global_licensing)
            # TODO: We create a copy of the project in the following
            # song-and-dance because the debian Copyright object cannot be
            # pickled.
            new_project = Project(
                project.root,
                vcs_strategy=project.vcs_strategy,
                license_map=project.license_map,
                licenses=project.licenses.copy(),
                # TODO: adjust this method/class to account for REUSE.toml as
                # well. Unset dep5_copyright
                global_licensing=None,
                include_submodules=project.include_submodules,
                include_meson_subprojects=project.include_meson_subprojects,
            )
            new_project.licenses_without_extension = (
                project.licenses_without_extension
            )
            self.project = new_project
        else:
            self.has_dep5 = False
            self.project = project

        self.reuse_dep5: ReuseDep5 | None = None
        self.do_checksum = do_checksum
        self.add_license_concluded = add_license_concluded

    def __call__(self, file_: StrPath) -> "_MultiprocessingResult":
        # By remembering that we've parsed the .reuse/dep5, we only parse it
        # once (the first time) inside of each process.
        if self.has_dep5 and not self.reuse_dep5:
            with contextlib.suppress(Exception):
                self.reuse_dep5 = ReuseDep5.from_file(
                    self.project.root / ".reuse/dep5"
                )
                self.project.global_licensing = self.reuse_dep5
        # pylint: disable=broad-except
        try:
            return _MultiprocessingResult(
                file_,
                FileReport.generate(
                    self.project,
                    file_,
                    do_checksum=self.do_checksum,
                    add_license_concluded=self.add_license_concluded,
                ),
                None,
            )
        except Exception as exc:
            return _MultiprocessingResult(file_, None, exc)


class _MultiprocessingResult(NamedTuple):
    """Result of :class:`MultiprocessingContainer`."""

    path: StrPath
    report: Optional["FileReport"]
    error: Exception | None


def _generate_file_reports(
    project: Project,
    do_checksum: bool = True,
    subset_files: Collection[StrPath] | None = None,
    multiprocessing: bool = _CPU_COUNT > 1,
    add_license_concluded: bool = False,
) -> Generator[_MultiprocessingResult, None, None]:
    """Create a :class:`FileReport` for every file in the project, filtered
    by *subset_files*.
    """
    container = _MultiprocessingContainer(
        project, do_checksum, add_license_concluded
    )

    files = (
        project.subset_files(subset_files)
        if subset_files is not None
        else project.all_files()
    )
    if multiprocessing and ENABLE_PARALLEL:
        files_set = frozenset(files)
        with ProcessPoolExecutor() as executor:
            yield from executor.map(
                container,
                files_set,
                chunksize=max(1, int(len(files_set) / _CPU_COUNT / 4)),
            )
    else:
        yield from map(container, files)


def _process_error(error: Exception, path: StrPath) -> None:
    # Facilitate better debugging by being able to quit the program.
    if isinstance(error, (bdb.BdbQuit, KeyboardInterrupt)):
        raise error
    if isinstance(error, (OSError, UnicodeError)):
        _LOGGER.error(
            _("Could not read '{path}'").format(path=path),
            exc_info=error,
        )
    else:
        _LOGGER.error(
            _("Unexpected error occurred while parsing '{path}'").format(
                path=path
            ),
            exc_info=error,
        )


class ProjectReportSubsetProtocol(Protocol):
    """A :class:`Protocol` that defines a subset of functionality of
    :class:`ProjectReport`, implemented by :class:`ProjectSubsetReport`.
    """

    path: StrPath
    read_errors: set[Path]
    file_reports: set["FileReport"]

    @property
    def missing_licenses(self) -> dict[str, set[Path]]:
        """Files which refer to a license which do not exist in the LICENSES/
        directory.
        """

    @property
    def invalid_spdx_expressions(self) -> dict[Path, set[str]]:
        """Invalid expressions by file."""

    @property
    def files_without_licenses(self) -> set[Path]:
        """Set of paths that have no licensing information."""

    @property
    def files_without_copyright(self) -> set[Path]:
        """Set of paths that have no copyright information."""

    @property
    def is_compliant(self) -> bool:
        """Whether the report subset is compliant with the REUSE Spec."""


class ProjectReport:
    """Object that holds linting report about the project."""

    def __init__(self, do_checksum: bool = True):
        self.path: StrPath = ""
        self.licenses: dict[str, Path] = {}
        self.read_errors: set[Path] = set()
        self.file_reports: set[FileReport] = set()
        self.licenses_without_extension: dict[str, Path] = {}

        self.do_checksum = do_checksum

        self._license_map: dict[str, dict] = {}

    def to_dict_lint(self) -> dict[str, Any]:
        """Collects and formats data relevant to linting from report and returns
        it as a dictionary.

        Returns:
            Dictionary containing data from the ProjectReport object.
        """
        # Setup report data container
        data: dict[str, Any] = {
            "non_compliant": {
                "bad_licenses": list(sorted(self.bad_licenses)),
                "deprecated_licenses": list(
                    sorted(str(file) for file in self.deprecated_licenses)
                ),
                "licenses_without_extension": list(
                    sorted(self.licenses_without_extension)
                ),
                "missing_licenses": list(sorted(self.missing_licenses)),
                "unused_licenses": list(
                    sorted(str(file) for file in self.unused_licenses)
                ),
                "read_errors": list(
                    sorted(str(file) for file in self.read_errors)
                ),
                "missing_copyright_info": list(
                    sorted(str(file) for file in self.files_without_copyright)
                ),
                "missing_licensing_info": list(
                    sorted(str(file) for file in self.files_without_licenses)
                ),
            },
            "files": [],
            "summary": {
                "used_licenses": [],
            },
            "recommendations": self.recommendations,
        }

        # Populate 'files'
        for file_report in self.file_reports:
            data["files"].append(file_report.to_dict_lint())

        # Populate 'summary'
        number_of_files = len(self.file_reports)
        data["summary"] = {
            "used_licenses": list(sorted(self.used_licenses)),
            "files_total": number_of_files,
            "files_with_copyright_info": number_of_files
            - len(self.files_without_copyright),
            "files_with_licensing_info": number_of_files
            - len(self.files_without_licenses),
            "compliant": self.is_compliant,
        }

        # Add the top three keys
        unsorted_data = {
            "lint_version": LINT_VERSION,
            "reuse_spec_version": __REUSE_version__,
            "reuse_tool_version": __version__,
            **data,
        }

        # Sort dictionary keys while keeping the top three keys at the beginning
        # and the recommendations on the bottom
        sorted_keys = sorted(list(unsorted_data.keys()))
        sorted_keys.remove("lint_version")
        sorted_keys.remove("reuse_spec_version")
        sorted_keys.remove("reuse_tool_version")
        sorted_keys.remove("recommendations")
        sorted_keys = (
            [
                "lint_version",
                "reuse_spec_version",
                "reuse_tool_version",
            ]
            + sorted_keys
            + ["recommendations"]
        )

        sorted_data = {key: unsorted_data[key] for key in sorted_keys}

        return sorted_data

    def bill_of_materials(
        self,
        creator_person: str | None = None,
        creator_organization: str | None = None,
    ) -> str:
        """Generate a bill of materials from the project.

        See https://spdx.org/specifications.
        """
        out = StringIO()
        # Write mandatory tags
        out.write("SPDXVersion: SPDX-2.1\n")
        out.write("DataLicense: CC0-1.0\n")
        out.write("SPDXID: SPDXRef-DOCUMENT\n")

        out.write(f"DocumentName: {Path(self.path).resolve().name}\n")
        # TODO: Generate UUID from git revision maybe
        # TODO: Fix the URL
        out.write(
            f"DocumentNamespace: http://spdx.org/spdxdocs/spdx-v2.1-{uuid4()}\n"
        )

        # Author
        out.write(f"Creator: Person: {format_creator(creator_person)}\n")
        out.write(
            f"Creator: Organization: {format_creator(creator_organization)}\n"
        )
        out.write(f"Creator: Tool: reuse-{__version__}\n")

        now = datetime.datetime.now(tz=datetime.timezone.utc)
        out.write(f"Created: {now.strftime('%Y-%m-%dT%H:%M:%SZ')}\n")
        out.write(
            "CreatorComment: <text>This document was created automatically"
            " using available reuse information consistent with"
            " REUSE.</text>\n"
        )

        reports = sorted(self.file_reports, key=lambda x: x.name)

        for report in reports:
            out.write(
                "Relationship: SPDXRef-DOCUMENT DESCRIBES"
                f" {report.spdx_id}\n"
            )

        for report in reports:
            out.write("\n")
            out.write(f"FileName: {report.name}\n")
            out.write(f"SPDXID: {report.spdx_id}\n")
            out.write(f"FileChecksum: SHA1: {report.chk_sum}\n")
            out.write(f"LicenseConcluded: {report.license_concluded}\n")

            for lic in sorted(report.licenses_in_file):
                out.write(f"LicenseInfoInFile: {lic}\n")
            if report.copyright:
                out.write(
                    "FileCopyrightText:" f" <text>{report.copyright}</text>\n"
                )
            else:
                out.write("FileCopyrightText: NONE\n")

        # Licenses
        for lic, path in sorted(self.licenses.items()):
            if _LICENSEREF_PATTERN.match(lic):
                out.write("\n")
                out.write(f"LicenseID: {lic}\n")
                out.write("LicenseName: NOASSERTION\n")

                with (Path(self.path) / path).open(encoding="utf-8") as fp:
                    out.write(f"ExtractedText: <text>{fp.read()}</text>\n")

        return out.getvalue()

    @classmethod
    def generate(
        cls,
        project: Project,
        do_checksum: bool = True,
        multiprocessing: bool = _CPU_COUNT > 1,
        add_license_concluded: bool = False,
    ) -> "ProjectReport":
        """Generate a :class:`ProjectReport` from a :class:`Project`.

        Args:
            project: The :class:`Project` to lint.
            do_checksum: Generate a checksum of every file. If this is
                :const:`False`, generate a random checksum for every file.
            multiprocessing: Whether to use multiprocessing.
            add_license_concluded: Whether to aggregate all found SPDX
                expressions into a concluded license.
        """
        project_report = cls(do_checksum=do_checksum)
        project_report.path = project.root
        project_report.licenses = project.licenses
        project_report._license_map = project.license_map
        project_report.licenses_without_extension = (
            project.licenses_without_extension
        )

        results = _generate_file_reports(
            project,
            do_checksum=do_checksum,
            multiprocessing=multiprocessing,
            add_license_concluded=add_license_concluded,
        )
        for result in results:
            if result.error:
                _process_error(result.error, result.path)
                project_report.read_errors.add(Path(result.path))
                continue

            file_report = cast(FileReport, result.report)
            project_report.file_reports.add(file_report)

        return project_report

    @cached_property
    def used_licenses(self) -> set[str]:
        """Set of license identifiers that are found in file reports."""
        return {
            lic
            for file_report in self.file_reports
            for lic in file_report.licenses_in_file
        }

    @cached_property
    def bad_licenses(self) -> dict[str, Path]:
        """Licenses in LICENSES/ which are not valid SPDX licenses."""
        return {
            lic: path
            for lic, path in self.licenses.items()
            if lic not in self._license_map
        }

    @cached_property
    def deprecated_licenses(self) -> set[str]:
        """Licenses whose SPDX License identifier has been deprecated."""
        return {
            lic
            for lic in self.licenses
            if lic in self._license_map
            and self._license_map[lic]["isDeprecatedLicenseId"]
        }

    @cached_property
    def unused_licenses(self) -> set[str]:
        """Set of license identifiers that are not found in any file report."""
        return {
            lic
            for lic in self.licenses
            if not any(
                identifier in self.used_licenses
                for identifier in (lic, _add_plus_to_identifier(lic))
            )
        }

    @cached_property
    def missing_licenses(self) -> dict[str, set[Path]]:
        """Files which refer to a license which do not exist in the LICENSES/
        directory.
        """
        result = defaultdict(set)
        for file_report in self.file_reports:
            for missing_license in file_report.missing_licenses:
                result[missing_license].add(file_report.path)
        return result

    @cached_property
    def invalid_spdx_expressions(self) -> dict[Path, set[str]]:
        """Invalid expressions by file."""
        return {
            file_report.path: file_report.invalid_spdx_expressions
            for file_report in self.file_reports
            if file_report.invalid_spdx_expressions
        }

    @cached_property
    def files_without_licenses(self) -> set[Path]:
        """Set of paths that have no licensing information."""
        return {
            file_report.path
            for file_report in self.file_reports
            if not file_report.licenses_in_file
        }

    @cached_property
    def files_without_copyright(self) -> set[Path]:
        """Set of paths that have no copyright information."""
        return {
            file_report.path
            for file_report in self.file_reports
            if not file_report.copyright
        }

    @cached_property
    def is_compliant(self) -> bool:
        """Whether the report is compliant with the REUSE Spec."""
        return not any(
            (
                self.missing_licenses,
                self.unused_licenses,
                self.bad_licenses,
                self.deprecated_licenses,
                self.licenses_without_extension,
                self.read_errors,
                self.invalid_spdx_expressions,
                self.files_without_copyright,
                self.files_without_licenses,
            )
        )

    @property
    def recommendations(self) -> list[str]:
        """Generate help for next steps based on found REUSE issues"""
        recommendations = []

        # These items should be ordered in the same way as in the summary.
        if self.bad_licenses:
            recommendations.append(
                _(
                    "Fix bad licenses: At least one license in the LICENSES"
                    " directory and/or provided by 'SPDX-License-Identifier'"
                    " tags is invalid. They are either not valid SPDX License"
                    " Identifiers or do not start with 'LicenseRef-'. FAQ about"
                    " custom licenses:"
                    " https://reuse.software/faq/#custom-license"
                )
            )
        if self.deprecated_licenses:
            recommendations.append(
                _(
                    "Fix deprecated licenses: At least one of the licenses in"
                    " the LICENSES directory and/or provided by an"
                    " 'SPDX-License-Identifier' tag or in '.reuse/dep5' has"
                    " been deprecated by SPDX. The current list and their"
                    " respective recommended  new identifiers can be found"
                    " here: <https://spdx.org/licenses/#deprecated>"
                )
            )
        if self.licenses_without_extension:
            recommendations.append(
                _(
                    "Fix licenses without file extension: At least one license"
                    " text file in the 'LICENSES' directory does not have a"
                    " '.txt' file extension. Please rename the file(s)"
                    " accordingly."
                )
            )
        if self.missing_licenses:
            recommendations.append(
                _(
                    "Fix missing licenses: For at least one of the license"
                    " identifiers provided by the 'SPDX-License-Identifier'"
                    " tags, there is no corresponding license text file in the"
                    " 'LICENSES' directory. For SPDX license identifiers, you"
                    " can simply run 'reuse download --all' to get any missing"
                    " ones. For custom licenses (starting with 'LicenseRef-'),"
                    " you need to add these files yourself."
                )
            )
        if self.unused_licenses:
            recommendations.append(
                _(
                    "Fix unused licenses: At least one of the license text"
                    " files in 'LICENSES' is not referenced by any file, e.g."
                    " by an 'SPDX-License-Identifier' tag. Please make sure"
                    " that you either tag the accordingly licensed files"
                    " properly, or delete the unused license text if you are"
                    " sure that no file or code snippet is licensed as such."
                )
            )
        if self.read_errors:
            recommendations.append(
                _(
                    "Fix read errors: At least one of the files in your"
                    " directory cannot be read by the tool. Please check the"
                    " file permissions. You will find the affected files at the"
                    " top of the output as part of the logged error messages."
                )
            )
        if self.invalid_spdx_expressions:
            recommendations.append(
                _(
                    "Fix invalid SPDX License Expressions: In one or more files"
                    " there are SPDX License Expressions which cannot be"
                    " parse. Check whether the value that follows"
                    " 'SPDX-License-Identifier:' is correct. If the detected"
                    " expression is not meant to be valid, put it between"
                    " 'REUSE-IgnoreStart' and 'REUSE-IgnoreEnd' comments."
                )
            )
        if self.files_without_copyright or self.files_without_licenses:
            recommendations.append(
                _(
                    "Fix missing copyright/licensing information: For one or"
                    " more files, the tool cannot find copyright and/or"
                    " licensing information. You typically do this by adding"
                    " 'SPDX-FileCopyrightText' and 'SPDX-License-Identifier'"
                    " tags to each file. The tutorial explains additional ways"
                    " to do this: <https://reuse.software/tutorial/>"
                )
            )

        return recommendations


class ProjectSubsetReport:
    """Like a :class:`ProjectReport`, but for a subset of the files using a
    subset of features.
    """

    def __init__(self) -> None:
        self.path: StrPath = ""
        self.read_errors: set[Path] = set()
        self.file_reports: set[FileReport] = set()

    @classmethod
    def generate(
        cls,
        project: Project,
        subset_files: Collection[StrPath],
        multiprocessing: bool = _CPU_COUNT > 1,
    ) -> "ProjectSubsetReport":
        """Generate a :class:`ProjectSubsetReport` from a :class:`Project`.

        Args:
            project: The :class:`Project` to lint.
            subset_files: Only lint the files in this list.
            multiprocessing: Whether to use multiprocessing.
        """
        subset_report = cls()
        subset_report.path = project.root
        results = _generate_file_reports(
            project,
            do_checksum=False,
            subset_files=subset_files,
            multiprocessing=multiprocessing,
            add_license_concluded=False,
        )
        for result in results:
            if result.error:
                _process_error(result.error, result.path)
                subset_report.read_errors.add(Path(result.path))
                continue

            file_report = cast(FileReport, result.report)
            subset_report.file_reports.add(file_report)

        return subset_report

    @property
    def missing_licenses(self) -> dict[str, set[Path]]:
        """Files which refer to a license which do not exist in the LICENSES/
        directory.
        """
        result = defaultdict(set)
        for file_report in self.file_reports:
            for missing_license in file_report.missing_licenses:
                result[missing_license].add(file_report.path)
        return result

    @property
    def invalid_spdx_expressions(self) -> dict[Path, set[str]]:
        """Invalid expressions by file."""
        return {
            file_report.path: file_report.invalid_spdx_expressions
            for file_report in self.file_reports
            if file_report.invalid_spdx_expressions
        }

    @property
    def files_without_licenses(self) -> set[Path]:
        """Set of paths that have no licensing information."""
        return {
            file_report.path
            for file_report in self.file_reports
            if not file_report.licenses_in_file
        }

    @property
    def files_without_copyright(self) -> set[Path]:
        """Set of paths that have no copyright information."""
        return {
            file_report.path
            for file_report in self.file_reports
            if not file_report.copyright
        }

    @property
    def is_compliant(self) -> bool:
        """Whether the report subset is compliant with the REUSE Spec."""
        return not any(
            (
                self.missing_licenses,
                self.files_without_copyright,
                self.files_without_licenses,
                self.read_errors,
            )
        )


class FileReport:  # pylint: disable=too-many-instance-attributes
    """Object that holds a linting report about a single file."""

    def __init__(self, name: str, path: StrPath, do_checksum: bool = True):
        self.name = name
        self.path = Path(path)
        self.do_checksum = do_checksum

        self.reuse_infos: list[ReuseInfo] = []

        self.spdx_id: str | None = None
        self.chk_sum: str | None = None
        self.licenses_in_file: list[str] = []
        self.license_concluded: str = ""
        self.copyright: str = ""

        self.missing_licenses: set[str] = set()
        self.invalid_spdx_expressions: set[str] = set()

    def to_dict_lint(self) -> dict[str, Any]:
        """Turn the report into a json-like dictionary with exclusively
        information relevant for linting.
        """
        return {
            "path": PurePath(self.name).as_posix(),
            "copyrights": [
                {
                    "value": str(line),
                    "source": reuse_info.source_path,
                    "source_type": (
                        reuse_info.source_type.value
                        if reuse_info.source_type
                        else None
                    ),
                }
                for reuse_info in self.reuse_infos
                for line in reuse_info.copyright_notices
            ],
            "spdx_expressions": [
                {
                    "value": str(expression),
                    "is_valid": expression.is_valid,
                    "source": reuse_info.source_path,
                    "source_type": (
                        reuse_info.source_type.value
                        if reuse_info.source_type
                        else None
                    ),
                }
                for reuse_info in self.reuse_infos
                for expression in reuse_info.spdx_expressions
            ],
        }

    @classmethod
    def generate(
        cls,
        project: Project,
        path: StrPath,
        do_checksum: bool = True,
        add_license_concluded: bool = False,
    ) -> "FileReport":
        """Generate a FileReport from a path in a Project."""
        # pylint: disable=too-many-branches
        path = Path(path)
        if not path.is_file():
            raise OSError(f"{path} is not a file")

        relative = project.relative_from_root(path)
        report = cls(f"./{relative}", path, do_checksum=do_checksum)

        # Checksum and ID
        if report.do_checksum:
            report.chk_sum = _checksum(path)
        else:
            # This path avoids a lot of heavy computation, which is handy for
            # scenarios where you only need a unique hash, not a consistent
            # hash.
            report.chk_sum = f"{random.getrandbits(160):040x}"
        spdx_id = md5()
        spdx_id.update(report.name.encode("utf-8"))
        spdx_id.update(report.chk_sum.encode("utf-8"))
        report.spdx_id = f"SPDXRef-{spdx_id.hexdigest()}"

        reuse_infos = project.reuse_info_of(path)
        for reuse_info in reuse_infos:
            for expression in reuse_info.spdx_expressions:
                if not expression.is_valid:
                    report.invalid_spdx_expressions.add(str(expression))
                    continue
                for identifier in expression.licenses:
                    # A license expression akin to Apache-1.0+ should register
                    # correctly if LICENSES/Apache-1.0.txt exists.
                    identifiers = {identifier}
                    if (
                        plus_identifier := _strip_plus_from_identifier(
                            identifier
                        )
                    ) != identifier:
                        identifiers.add(plus_identifier)
                    # Missing license
                    if not identifiers.intersection(project.licenses):
                        report.missing_licenses.add(identifier)

                    # Add license to report.
                    report.licenses_in_file.append(identifier)

        if not add_license_concluded:
            report.license_concluded = "NOASSERTION"
        elif not any(reuse_info.spdx_expressions for reuse_info in reuse_infos):
            report.license_concluded = "NONE"
        elif report.invalid_spdx_expressions:
            report.license_concluded = "NOASSERTION"
        else:
            # Merge all the license expressions together, wrapping them in
            # parentheses to make sure an expression doesn't spill into another
            # one. The extra parentheses will be removed by the roundtrip
            # through parse() -> simplify() -> render().
            report.license_concluded = str(
                SpdxExpression.combine(
                    list(
                        expression
                        for reuse_info in reuse_infos
                        for expression in reuse_info.spdx_expressions
                    )
                ).simplify()
            )

        # Copyright text
        report.copyright = "\n".join(
            map(
                str,
                sorted(
                    line
                    for reuse_info in reuse_infos
                    for line in reuse_info.copyright_notices
                ),
            )
        )
        # Source of licensing and copyright info
        report.reuse_infos = reuse_infos
        return report

    def __hash__(self) -> int:
        if self.chk_sum is not None:
            return hash(self.name + self.chk_sum)
        return super().__hash__()


def format_creator(creator: str | None) -> str:
    """Render the creator field based on the provided flag"""
    if creator is None:
        return "Anonymous ()"
    if "(" in creator and creator.endswith(")"):
        # The creator field already contains an email address
        return creator
    return creator + " ()"


# REUSE-IgnoreEnd



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2024 Carmen Bianca BAKKER <carmenbianca@fsfe.org>
# SPDX-FileCopyrightText: © 2020 Liferay, Inc. <https://liferay.com>
# SPDX-FileCopyrightText: 2024 Kerry McAdams <github@klmcadams>
# SPDX-FileCopyrightText: 2024 Emil Velikov <emil.l.velikov@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Entry function for reuse."""

import gettext
import logging
import os
import warnings
from pathlib import Path

import click
from click.formatting import wrap_text

from .. import __REUSE_version__
from .._util import setup_logging
from ..i18n import _
from .common import ClickObj

_PACKAGE_PATH = os.path.dirname(os.path.dirname(__file__))
_LOCALE_DIR = os.path.join(_PACKAGE_PATH, "locale")
if gettext.find("reuse", localedir=_LOCALE_DIR):
    gettext.bindtextdomain("reuse", _LOCALE_DIR)
    # This is needed to make Click recognise our translations. Our own
    # translations use the class-based API.
    gettext.textdomain("reuse")


_VERSION_TEXT = (
    _("%(prog)s, version %(version)s")
    + "\n\n"
    + _(
        "This program is free software: you can redistribute it and/or modify"
        " it under the terms of the GNU General Public License as published by"
        " the Free Software Foundation, either version 3 of the License, or"
        " (at your option) any later version."
    )
    + "\n\n"
    + _(
        "This program is distributed in the hope that it will be useful,"
        " but WITHOUT ANY WARRANTY; without even the implied warranty of"
        " MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the"
        " GNU General Public License for more details."
    )
    + "\n\n"
    + _(
        "You should have received a copy of the GNU General Public License"
        " along with this program. If not, see"
        " <https://www.gnu.org/licenses/>."
    )
)

_HELP = (
    _(
        "reuse is a tool for compliance with the REUSE"
        " recommendations. See <https://reuse.software/> for more"
        " information, and <https://reuse.readthedocs.io/> for the online"
        " documentation."
    )
    + "\n\n"
    + _(
        "This version of reuse is compatible with version {} of the REUSE"
        " Specification."
    ).format(__REUSE_version__)
    + "\n\n"
    + _("Support the FSFE's work:")
    + "\n\n"
    # Indent next paragraph.
    + "   "
    + _(
        "Donations are critical to our strength and autonomy. They enable us"
        " to continue working for Free Software wherever necessary. Please"
        " consider making a donation at <https://fsfe.org/donate/>."
    )
)


@click.group(name="reuse", help=_HELP)
@click.option(
    "--debug",
    is_flag=True,
    help=_("Enable debug statements."),
)
@click.option(
    "--suppress-deprecation",
    is_flag=True,
    help=_("Hide deprecation warnings."),
)
@click.option(
    "--include-submodules",
    is_flag=True,
    help=_("Do not skip over Git submodules."),
)
@click.option(
    "--include-meson-subprojects",
    is_flag=True,
    help=_("Do not skip over Meson subprojects."),
)
@click.option(
    "--no-multiprocessing",
    is_flag=True,
    help=_("Do not use multiprocessing."),
)
@click.option(
    "--root",
    type=click.Path(
        exists=True,
        file_okay=False,
        path_type=Path,
    ),
    default=None,
    help=_("Define root of project."),
)
@click.version_option(
    package_name="reuse",
    message=wrap_text(_VERSION_TEXT, preserve_paragraphs=True),
)
@click.pass_context
def main(
    ctx: click.Context,
    debug: bool,
    suppress_deprecation: bool,
    include_submodules: bool,
    include_meson_subprojects: bool,
    no_multiprocessing: bool,
    root: Path | None,
) -> None:
    # pylint: disable=missing-function-docstring,too-many-arguments
    setup_logging(level=logging.DEBUG if debug else logging.WARNING)

    # Very stupid workaround to not print a DEP5 deprecation warning in the
    # middle of ccompileonversion to REUSE.toml.
    if ctx.invoked_subcommand == "convert-dep5":
        os.environ["_SUPPRESS_DEP5_WARNING"] = "1"

    if not suppress_deprecation:
        warnings.filterwarnings("default", module="reuse")

    ctx.obj = ClickObj(
        root=root,
        include_submodules=include_submodules,
        include_meson_subprojects=include_meson_subprojects,
        no_multiprocessing=no_multiprocessing,
    )



# SPDX-FileCopyrightText: 2024 Carmen Bianca BAKKER <carmen@carmenbianca.eu>
# SPDX-FileCopyrightText: 2024 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

""":mod:`gettext` plumbing of :mod:`reuse`."""

import gettext as _gettext_module
import os

_PACKAGE_PATH = os.path.dirname(__file__)
_LOCALE_DIR = os.path.join(_PACKAGE_PATH, "locale")

#: Translations object used throughout :mod:`reuse`. The translations
#: are sourced from ``reuse/locale/<lang>/LC_MESSAGES/reuse.mo``.
TRANSLATIONS: _gettext_module.NullTranslations = _gettext_module.translation(
    "reuse", localedir=_LOCALE_DIR, fallback=True
)
#: :meth:`gettext.NullTranslations.gettext` of :data:`TRANSLATIONS`
_ = TRANSLATIONS.gettext
#: :meth:`gettext.NullTranslations.gettext` of :data:`TRANSLATIONS`
gettext = TRANSLATIONS.gettext
#: :meth:`gettext.NullTranslations.ngettext` of :data:`TRANSLATIONS`
ngettext = TRANSLATIONS.ngettext
#: :meth:`gettext.NullTranslations.pgettext` of :data:`TRANSLATIONS`
pgettext = TRANSLATIONS.pgettext
#: :meth:`gettext.NullTranslations.npgettext` of :data:`TRANSLATIONS`
npgettext = TRANSLATIONS.npgettext



# SPDX-FileCopyrightText: 2021 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2021 Michael Weimann
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2025 Shun Sakai <sorairolake@protonmail.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Click code for supported-licenses subcommand."""

import json

import click

from .._licenses import _LICENSES, _load_license_list
from ..i18n import _
from .common import ClickObj
from .main import main

_HELP = _("List all licenses on the SPDX License List.")


@main.command(name="supported-licenses", help=_HELP)
@click.option(
    "--json",
    "-j",
    "format_json",
    is_flag=True,
    help=_("Format output as JSON."),
)
@click.pass_obj
def supported_licenses(_obj: ClickObj, format_json: bool) -> None:
    # pylint: disable=missing-function-docstring
    licenses = _load_license_list(_LICENSES)[1]

    if format_json:
        result = [
            {
                "id": license_id,
                "name": license_info["name"],
                "reference": license_info["reference"],
            }
            for license_id, license_info in licenses.items()
        ]
        click.echo(json.dumps(result, indent=2), nl=False)
    else:
        for license_id, license_info in licenses.items():
            license_name = license_info["name"]
            license_reference = license_info["reference"]
            click.echo(
                f"{license_id: <40}\t{license_name: <80}\t"
                f"{license_reference: <50}"
            )
